/******************************************************************************
 * @file chan_hss.c
 *
 * @description HSS Channel driver for Asterisk
 *
 *
 * @Revision 1.0
 * This file is provided under a dual BSD/GPLv2 license.  When using or 
 *   redistributing this file, you may do so under either license.
 * 
 *   GPL LICENSE SUMMARY
 * 
 *   Copyright(c) 2007,2008 Intel Corporation. All rights reserved.
 * 
 *   This program is free software; you can redistribute it and/or modify 
 *   it under the terms of version 2 of the GNU General Public License as
 *   published by the Free Software Foundation.
 * 
 *   This program is distributed in the hope that it will be useful, but 
 *   WITHOUT ANY WARRANTY; without even the implied warranty of 
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
 *   General Public License for more details.
 * 
 *   You should have received a copy of the GNU General Public License 
 *   along with this program; if not, write to the Free Software 
 *   Foundation, Inc., 51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA.
 *   The full GNU General Public License is included in this distribution 
 *   in the file called LICENSE.GPL.
 * 
 *   Contact Information:
 *   Intel Corporation
 * 
 *   BSD LICENSE 
 * 
 *   Copyright(c) 2007,2008 Intel Corporation. All rights reserved.
 *   All rights reserved.
 * 
 *   Redistribution and use in source and binary forms, with or without 
 *   modification, are permitted provided that the following conditions 
 *   are met:
 * 
 *     * Redistributions of source code must retain the above copyright 
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright 
 *       notice, this list of conditions and the following disclaimer in 
 *       the documentation and/or other materials provided with the 
 *       distribution.
 *     * Neither the name of Intel Corporation nor the names of its 
 *       contributors may be used to endorse or promote products derived 
 *       from this software without specific prior written permission.
 * 
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 *   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
 *   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
 *   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 *   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 *   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
 *   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
 *   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 *   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * 
 *
 *****************************************************************************/


/*!
 * \file
 * \brief HSS Channel driver for Asterisk
 */

#include <stdio.h>
#include <pthread.h>
#include <string.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/poll.h>

#include "asterisk/config.h"
#include "asterisk/channel.h"
#include "asterisk/callerid.h"
#include "asterisk/logger.h"
#include "asterisk/module.h"
#include "asterisk/options.h"
#include "asterisk/causes.h"
#include "asterisk/pbx.h"
#include "asterisk/dsp.h"
#include "asterisk/translate.h"
#include "asterisk/musiconhold.h"
#include "asterisk/ulaw.h"
#include "asterisk/alaw.h"
#include "asterisk/cli.h"

#include "icp.h"
#include "icp_hssvoicedrv.h"
#include "icp_hssdrv.h"
#include "icp_analogdrv.h"
#include "icp_framerdrv.h"

#ifdef USE_OSLEC_EC
#include "spandsp/echo.h"
#endif

#define DESCRIPTION "HSS Channel Driver for Asterisk"
#define CONFIG_FILE "hss.conf"
#define VOICE_DEVICE "/dev/hss-voice"
#define SLIC_DEVICE "/dev/analog-fxo-fxs"



#define MAX_FXSFXO_NUM     15    /*!< Number of FXS/FXO lines supported */

#define MAX_DIGITAL_NUM    120    /*!< Number of DIGITAL timeslots supported */

#define MAX_CHANNELS_NUM   128    /*!< Number of channels supported */


#define NB_HSS_PORTS       3
#define NB_SPANS_PER_MEZZANINE    4

#define US_FCC      0
#define EU_TBR21    1

/*!< priority of monitor threads, both analog and digital */
#define MONITOR_THREAD_PRIORITY    30

/*!< priority of threads reading the digits */
#define DIGITS_THREAD_PRIORITY     20


/*!< Maximum on hook time allowed for hookflash event (in ms) */
#define HOOKFLASHTIME   500    

/*!< digit timeout used when the user is dialing over FXO/FXS (in ms) */
#define ANALOGDIGITTIMEOUT    2000

/*!< digit timeout used when the PBX is dialing over DIGITAL interface (in ms) */
#define DIGITALDIGITTIMEOUT    500

/*!< silence length between 2 digits when sent over FXO (in ms) */
#define INTERDIGITTIME  100

#define TAPS     256            /*!< Tail length of the echo canceller in bits */

/*!< We use a circular buffer to store voice samples required for performing the echo
     cancellation. It is made up of 8 buffers, each containing 80 bytes. This allows to
     support different read rates: 10, 20, 30 and 40ms for G711 and 10 or 20ms for linear.
      */
#define NB_EC_BUFFERS    8


#define DIGITAL_LINE_MASK     10000
#define SLOT_MASK              1000
#define SPAN_MASK               100


/*!< #define's used for the voice driver */
/*!< header size of a voice packet in bytes */
#define VOICE_HEADER_SIZE      4    

/*!< size of a voice packet in 8kHz voice samples */
#define VOICE_PAYLOAD_SIZE    80    

/*!< size of an ulaw/alaw voice packet retrieved by the voice driver, in bytes */
#define VOICE_ULAW_PACKET_SIZE     84    

/*!< size of a linear voice packet retrieved by the voice driver, in bytes */
#define VOICE_LINEAR_PACKET_SIZE     164    


#define MAX_TIMESLOT    32

#define POLL_TIMEOUT    500

/*!< #define's used for the SLIC/CODEC driver */
#define SIZE_OF_BYTE    8

#define CLK_SPEED_2MHZ    0

#define PCM_ULAW    0
#define PCM_ALAW    1
#define PCM_LINEAR  2

#define FXS_OFFHOOK     1
#define FXS_ONHOOK      0

#define TONE_REVERSE_DIRECTION 1

/*!< settings of the asterisk DSP */
#define SILENCE_THRESHOLD    128
#define BUSY_PATTERN         500
#define BUSY_COUNT             4

/*!< Macro to configure a voice channel */
#define CONFIG_CHANNEL(chan) do {     \
    chan.channelId = 0;               \
    chan.voiceIdleAction = 0;         \
    chan.channelDataInvert = 0;       \
    chan.channelBitEndianness = 1;    \
    chan.channelByteSwap = 0;         \
    } while(0);

    
/*!< Macro to bring up a voice channel */
#define BRING_CHANNEL_UP(pvt) do {                                       \
    ast_mutex_lock(&pvt->lock);                                          \
    if (pvt->up == FALSE) {                                              \
        if (ioctl(pvt->fd, ICP_HSSVOICEDRV_CHAN_UP, 0)) {                \
            ast_log (LOG_ERROR, "Fail to bring up the voice channel\n"); \
        } else {                                                         \
            pvt->up = TRUE;                                              \
        }                                                                \
    }                                                                    \
    ast_mutex_unlock(&pvt->lock);                                        \
} while(0);

/*!< Macro to bring down a voice channel */
#define BRING_CHANNEL_DOWN(pvt) do {                                       \
    ast_mutex_lock(&pvt->lock);                                            \
    if (pvt->up == TRUE) {                                                 \
        if (ioctl(pvt->fd, ICP_HSSVOICEDRV_CHAN_DOWN, 0)) {                \
            ast_log (LOG_ERROR, "Fail to bring down the voice channel\n"); \
        } else {                                                           \
            pvt->up = FALSE;                                               \
        }                                                                  \
    }                                                                      \
    ast_mutex_unlock(&pvt->lock);                                          \
} while(0);

/*!< Macro to swap bytes. Used only with linear codec */
#define SWAP_BYTES(a, b) do {         \
             unsigned char temp = b;  \
             b = a;                   \
             a = temp;                \
             }while(0);

#define EM_ONHOOK     0x01
#define EM_OFFHOOK    0x0D

/******************************************************************/
/* Types definitions */
/******************************************************************/

/*! \brief Type of interface */
typedef enum {
    FXS = 0,
    FXO,
    DIGITAL
} chan_mode;


/*! \brief Type of call */
typedef enum {
    MAINCALL = 0,
    WAITINGCALL
} call_type;


/*! \brief Configuration mode of an HSS port */
typedef enum {
    SLIC = 0,                    /*!< Analog mezzanine */
    E1,                          /*!< T1/E1 mezzanine in E1 mode (single or quad)*/
    T1                           /*!< T1/E1 mezzanine in T1 mode (single or quad)*/
} port_mode;


/*! \brief Setting for the E1/T1 mezzanine, single or quad configuration */
typedef enum {
    SINGLE = 0,
    QUAD
} port_setting;


/*! \brief Configuration parameters of an HSS port */
struct port {
    icp_hssdrv_portup_t config;
    unsigned int configured;    /*!< Flag indicating whether the port has been configured or not */
    unsigned int up;            /*!< Flag indicating whether the port is up or not */
    port_mode mode;
    port_setting setting;
    pthread_t monitorThread;    /*!< Identifier of the monitor thread associated to this port (DIGITAL only) */
    unsigned int availableTimeslot;    /*!< Next available timeslot on this port (used only by FXS/FXO) */
};


/*! \brief Private structure used for each interface (FXS, FXO or timeslot for DIGITAL interfaces)*/
struct hss_pvt {
    ast_mutex_t lock;              /*!< mutex to protect the structure */
    int fd;                        /*!< File descriptor to read/write voice data */
    struct ast_channel *owner[2];  /*!< Logical channels for main call and waiting call */
    struct ast_frame inFrame;
    struct ast_dsp *dsp;        /*!< DSP to perform digit detection */
    unsigned int slotID;        /*!< Hss port we are plugged in */
    unsigned int spanID;        /*!< Span ID (DIGITAL only) */
    unsigned int lineID;        /*!< Line ID */
    pthread_t getDigits_thread; /*!< Thread retrieving the digits */
    pthread_t timer_thread;     /*!< Thread used to handle an onhook event and detect hookflash */
    unsigned int needhangup;    /*!< flag to notify if line is off hook or not */
    unsigned char rxBuffer[VOICE_LINEAR_PACKET_SIZE];    /*!< Buffer to read data */
    unsigned char txBuffer[NB_EC_BUFFERS][VOICE_PAYLOAD_SIZE];    /*!< Buffer to store data received from asterisk */
    unsigned int indexWrite, indexRead;
    char ext[AST_MAX_EXTENSION];
    char context[AST_MAX_EXTENSION];
    char cid_num[AST_MAX_EXTENSION];
    char cid_name[AST_MAX_EXTENSION];
    int format;
    chan_mode mode;
    int timeslot;                /*!< Timeslot number, values: 0 to 31 */
    struct timeval onhooktime;    /*!< Timer used to determine hookflash */
    unsigned int callwaiting;    /*!< Flag to activate call waiting feature */
    unsigned int echocancel;    /*!< Flag to activate echo cancellation */
    unsigned int up;            /*!< Flag indicating whether the channel is up or not */
#ifdef USE_OSLEC_EC
    echo_can_state_t *ec;        /*!< Echo canceller */
#endif
};


/******************************************************************/
/* Global variables */
/******************************************************************/
/*!< Monitor thread for FXS/FXO interfaces */
static pthread_t monitor_thread = AST_PTHREADT_NULL; 
   
static int usecnt = 0;    /*!< Number of channels currently in use */
static int slicfd = -1;   /*!< File descriptor to read events from the SLIC/CODEC */
AST_MUTEX_DEFINE_STATIC (iflock);    /*!< Mutex to protect the interface array */


/* Global variables set when reading the configuration file */
static int configFormat = AST_FORMAT_ULAW;
static char callerid_name[AST_MAX_EXTENSION];
static char callerid_num[AST_MAX_EXTENSION];
static char context[AST_MAX_EXTENSION] = "default";
static int silencesuppression = FALSE;
static int callwaiting = FALSE;
static int echocancel = FALSE;


static struct port hssPorts[NB_HSS_PORTS];
static struct hss_pvt interfaces[MAX_CHANNELS_NUM];

static unsigned int nbInterfaces = 0;    /*!< Number of configured interfaces */
static unsigned int nbFxsFxo = 0;    /*!< Number of configured FXS/FXO interfaces */
static unsigned int nbDigital = 0;    /*!< Number of configured DIGITAL timeslots */



/******************************************************************/
/* Functions prototypes */
/******************************************************************/
/* Callbacks required by Asterisk */
static int hss_call (struct ast_channel *chan, char *dest, int timeout);
static int hss_hangup (struct ast_channel *chan);
static int hss_answer (struct ast_channel *chan);
static int hss_write (struct ast_channel *chan, struct ast_frame *frame);
static struct ast_frame *hss_read (struct ast_channel *chan);
static struct ast_channel *hss_request (const char *type, int format, void *data,
                                        int *cause);
static int hss_fixup (struct ast_channel *old, struct ast_channel *new);
static int hss_indicate (struct ast_channel *chan, int condition, const void *data,
                         size_t datalen);
static int hss_digit_begin (struct ast_channel *chan, char digit);
static int hss_digit_end (struct ast_channel *chan, char digit, unsigned int duration);
static int handle_hss_framer_status (int fd, int argc, char *argv[]);

/* Internal functions */
static int __unload_module (void);
static int setup_pcm (struct hss_pvt *pvt);
static int setup_digital (struct hss_pvt *pvt);
static struct ast_channel *newChannel (struct hss_pvt *pvt, call_type index, int state,
                                       char *context);
static void switch_channels (struct hss_pvt *pvt);
static int configure_port (unsigned int portID, port_mode mode, port_setting setting,
                           unsigned int preconfig, unsigned int geoconfig);
static struct hss_pvt *find_line (unsigned int id);
static struct hss_pvt *find_available_line_on_span (unsigned int slot, unsigned int span);
static int build_channels (char *v);
static int toInt(uint32_t n);
static void set_priority(int prio);

/*! \brief Definition of HSS channel type for Asterisk registration */
static const struct ast_channel_tech hss_tech = {
    .type = "HSS",
    .description = DESCRIPTION,
    .capabilities = AST_FORMAT_ALAW | AST_FORMAT_ULAW | AST_FORMAT_SLINEAR,
    .requester = hss_request,
    .call = hss_call,
    .hangup = hss_hangup,
    .answer = hss_answer,
    .read = hss_read,
    .write = hss_write,
    .fixup = hss_fixup,
    .indicate = hss_indicate,
    .send_digit_begin = hss_digit_begin,
    .send_digit_end = hss_digit_end
};



/* Variables required to register the CLI command printing out the framer driver status */
static char hss_framer_status_usage[] = 
    "Usage: hss framer status\n"
    "       Prints out the current state of each supported device\n";

static struct ast_cli_entry hss_cli = {{"hss", "framer", "status", NULL}, 
                                        handle_hss_framer_status,
                                        "Show framer driver status",
                                        hss_framer_status_usage};


                                        
/******************************************************************/
/*! \brief Convert a bitmap timeslot to an integer (2**x -> x)
    \param n bitmap to convert
    \return integer corresponding to the bit set to 1 */
/******************************************************************/
int toInt(uint32_t n)
{
    int val = 0;
    
    while(! (n & 1))
    {
        val++;
        n = n >> 1;
    }
    
    return val;
}


/******************************************************************/
/*! \brief Switch between the main call and the waiting call.       
           Used only when callwaiting is enabled
    \param pvt Interface to perform the switch on */
/******************************************************************/
void switch_channels (struct hss_pvt *pvt)
{
    struct ast_channel *temp = NULL;

    temp = pvt->owner[MAINCALL];
    pvt->owner[MAINCALL] = pvt->owner[WAITINGCALL];
    pvt->owner[WAITINGCALL] = temp;

    pvt->owner[MAINCALL]->fds[0] = pvt->fd;

    if (pvt->owner[WAITINGCALL] != NULL) {
        pvt->owner[WAITINGCALL]->fds[0] = -1;
    }

}

/******************************************************************/
/*! \brief Change priority of the thread calling that function. 
    \param prio priority to set the thread to */
/******************************************************************/
void set_priority (int prio)
{
    struct sched_param param;
    param.sched_priority = prio;
    
    if (pthread_setschedparam(pthread_self(), SCHED_RR, &param)) {
        /* Can't change the priority, thread will run with its default priority */
        ast_log (LOG_ERROR, "can't set priority!\n");
        
    } else {
        ast_log (LOG_DEBUG, "priority set to %d\n", prio);
    }
}


/******************************************************************/
/*! \brief Search for a line ID in the interfaces array
    \param id Line ID to search for 
    \return Pointer to the hss_pvt structure else NULL */
/******************************************************************/
struct hss_pvt *find_line (unsigned int id)
{
    int i = 0;

    for (i = 0; i < nbInterfaces; i++) {
        if (interfaces[i].lineID == id) {
            ast_log (LOG_DEBUG, "Found interface %d\n", id);
            return &interfaces[i];
        }
    }

    
    return NULL;
}

/******************************************************************/
/*! \brief Search for an available line over a span
    \param slot HSS port ID
    \param span Span ID on this port
    \return Pointer to the hss_pvt structure else NULL */
/******************************************************************/
struct hss_pvt *find_available_line_on_span (unsigned int slot, unsigned int span)
{
    int i = 0;
    struct hss_pvt *ret = NULL;
    uint8_t state = 0;
    icp_status_t status = 0;

    if (slot >= NB_HSS_PORTS || span >= NB_SPANS_PER_MEZZANINE) {
        ast_log (LOG_ERROR, "Invalid slot %d or span %d\n", slot, span);
        return NULL;
    }


/* Each timeslot is bound to an hss_pvt structure and is seen as a line 
   from asterisk perspective. We check every timeslots until we find an available one */
    for (i = 0; i < nbInterfaces; i++) {
        if (interfaces[i].mode == DIGITAL) {
            if ((interfaces[i].slotID == slot) && (interfaces[i].spanID == span)) {
            
                status = icp_FramerDrvSignalStateGet (slot, span, 
                                                      interfaces[i].timeslot, &state);

                if (status != ICP_STATUS_SUCCESS) {
                    continue;
                }

                /* Check if the line is available */
                if ((interfaces[i].owner[MAINCALL] == NULL) && (state == EM_ONHOOK)) {
                    ret = &interfaces[i];
                    ast_log (LOG_DEBUG, "Found line %d\n", ret->lineID);
                    break;
                }
            }
        }
    }


    return ret;
}


/******************************************************************/
/*! \brief Open a file descriptor and configure codecs and 
           timeslots for an FXO/FXS interface
    \param pvt Interface to configure
    \return 0 on success else -1 */
/******************************************************************/
static int setup_pcm (struct hss_pvt *pvt)
{
    icp_hssvoicedrv_channeladd_t hssChannel = {0};
    icp_analogdrv_pcm_config_t config = {0};
    icp_hssdrv_timeslot_map_t map = { 0, 0, 0, 0 };
    int tsAddedSuccessfully = -1;


    /* Try to add a channel until an available timeslot is found */
    while ((tsAddedSuccessfully != 0) &&
           (hssPorts[pvt->slotID].availableTimeslot < MAX_TIMESLOT)) {

        pvt->timeslot = hssPorts[pvt->slotID].availableTimeslot;
        hssPorts[pvt->slotID].availableTimeslot += 1;

        map.line0_timeslot_bit_map = (1 << pvt->timeslot);

        switch (pvt->format) {
        case AST_FORMAT_SLINEAR:
            if (pvt->timeslot == (MAX_TIMESLOT - 1)) {
                ast_log (LOG_ERROR, "Not enough timeslots to add a linear channel\n");
                return -1;
            }

            config.pcm_format = PCM_LINEAR;
            config.pcm_width = (SIZE_OF_BYTE * 2);
            hssChannel.voicePacketSize = (VOICE_PAYLOAD_SIZE * 2);
            map.line0_timeslot_bit_map |= (map.line0_timeslot_bit_map << 1);
            hssChannel.voiceIdlePattern = 0x00;    /* Silence */
            break;

        case AST_FORMAT_ULAW:
            config.pcm_format = PCM_ULAW;
            config.pcm_width = SIZE_OF_BYTE;
            hssChannel.voicePacketSize = VOICE_PAYLOAD_SIZE;
            hssChannel.voiceIdlePattern = 0xFF;    /* Silence */
            break;

        case AST_FORMAT_ALAW:
            config.pcm_format = PCM_ALAW;
            config.pcm_width = SIZE_OF_BYTE;
            hssChannel.voicePacketSize = VOICE_PAYLOAD_SIZE;
            hssChannel.voiceIdlePattern = 0x55;    /* Silence */
            break;

        default:
            ast_log (LOG_ERROR, "Codec not supported\n");
            return -1;
        }
   
        CONFIG_CHANNEL (hssChannel);
        hssChannel.portId = pvt->slotID;
        hssChannel.tsMap = map;

       
        /* We now try to add this voice channel. If ioctl fails, we'll go for another try
           with the next available timeslot until there's no free timeslot left */
        tsAddedSuccessfully = ioctl (pvt->fd, ICP_HSSVOICEDRV_CHAN_ADD, &hssChannel);
    }

    /* Can't add a voice channel */
    if (hssPorts[pvt->slotID].availableTimeslot >= MAX_TIMESLOT) {
        ast_log (LOG_ERROR, "Can't find an available timeslot\n");
        return -1;
    }

    if (config.pcm_width == (SIZE_OF_BYTE * 2)) {
        hssPorts[pvt->slotID].availableTimeslot += 1;
    }

    config.lineId = pvt->lineID;
    config.pcm_samp_freq = 8;
    config.pcm_start_offset = (pvt->timeslot * SIZE_OF_BYTE);

    /*Configure the SLIC/CODEC */
    if (ioctl (slicfd, ICP_ANALOGDRV_PCM_CONFIG_SET, &config)) {
        ast_log (LOG_ERROR, "Error in PCM configuration\n");
        return -1;
    }

    return 0;
}



/******************************************************************/
/*! \brief Open a file descriptor and configure timeslot for
           a DIGITAL interface
   \param pvt Interface to configure
   \return 0 on success else -1 */
/******************************************************************/
static int setup_digital (struct hss_pvt *pvt)
{

    icp_hssdrv_timeslot_map_t map = { 0, 0, 0, 0 };
    icp_hssvoicedrv_channeladd_t hssChannel;
    icp_status_t ret = ICP_STATUS_FAIL;
    unsigned int timeslot = 0;

    /*This is different from the analog scenario because in the case of digital lines
      each timeslot is bound to an hss_pvt structure. When chan_hss initialises the
      interfaces array and calls this function, each hss_pvt structure is already set
      to a specific timeslot so it just allocates the voice channel for this timeslot.
      In the case of analog lines, there is no predefined timeslot and that's why
      we need to look for an available one.
      Depending on the type of line, T1 or E1, the timeslot map is different (refer to
      icp_framerdrv.h for further details). */
   
    if (hssPorts[pvt->slotID].mode == E1) {
        
        timeslot = pvt->timeslot;
        
    } else  if (hssPorts[pvt->slotID].mode == T1) {
        
        /* The backplane is always set to E1, so when the framer driver is configured
           for T1 we need to map the 24 T1 timeslots to the 32 E1 timeslots according
           to the timeslot map provided in icp_framerdrv.h */
        timeslot = pvt->timeslot + ((pvt->timeslot - 1) / 3);
 
        if (timeslot >= MAX_TIMESLOT) {
            ast_log (LOG_ERROR, "Bad timeslot number. "
                     "Mode is set to T1, check preconfiguration and line IDs\n");
            return -1;
        }
    }
      
      
    switch (pvt->spanID) {
    case 0:
        map.line0_timeslot_bit_map = (1 << timeslot);
        break;

    case 1:
        map.line1_timeslot_bit_map = (1 << timeslot);
        break;

    case 2:
        map.line2_timeslot_bit_map = (1 << timeslot);
        break;

    case 3:
        map.line3_timeslot_bit_map = (1 << timeslot);
        break;

    default:
        ast_log (LOG_ERROR, "Invalid span ID\n");
        return -1;
        break;
    }

    CONFIG_CHANNEL (hssChannel);
    hssChannel.voicePacketSize = VOICE_PAYLOAD_SIZE;
    hssChannel.portId = pvt->slotID;
    hssChannel.tsMap = map;

    /*Setting silence pattern */
    if (pvt->format == AST_FORMAT_ULAW) {
        hssChannel.voiceIdlePattern = 0xFF;
    } else if (pvt->format == AST_FORMAT_ALAW) {
        hssChannel.voiceIdlePattern = 0x55;
    } else if (pvt->format == AST_FORMAT_SLINEAR) {
        ast_log (LOG_ERROR, "Linear not supported on digital lines\n");
        return -1;
    }


    /* Add a channel to the voice driver */
    if (ioctl (pvt->fd, ICP_HSSVOICEDRV_CHAN_ADD, &hssChannel)) {
        ast_log (LOG_ERROR, "Can't add a voice channel\n");
        return -1;
    }

    ret = icp_FramerDrvNotificationsSet (pvt->slotID, pvt->spanID,
                                         ICP_FRAMERDRV_EVENT_COSS);

    if (ret != ICP_STATUS_SUCCESS) {
        ast_log (LOG_ERROR, "Can't set the notifications for the Framer driver\n");
        return -1;
    }
    
    /* Set the signal so the timeslot is on hook */
    ret = icp_FramerDrvSignalStateSet (pvt->slotID, pvt->spanID, pvt->timeslot,
                                       EM_ONHOOK);

    if (ret != ICP_STATUS_SUCCESS) {
        ast_log (LOG_ERROR, 
                 "icp_FramerDrvSignalSet failed on slot %d, span %d, timeslot %d\n",
                 pvt->slotID, pvt->spanID, pvt->timeslot);
            return -1;
        }

    return 0;
}


/******************************************************************/
/*! \brief Configure an interface (intialisation, set parameters, 
           configure the Voice driver and the SLIC/CODEC driver)
    \param pos Index in the global array of interfaces
    \param lineID Line ID to add to the interfaces array 
    \return 0 on success else -1 */
/******************************************************************/
static int mkif (int pos, unsigned int lineID)
{

    int features = DSP_FEATURE_DTMF_DETECT;

    /* max limit of interfaces reached */
    if (pos == MAX_CHANNELS_NUM) {
        ast_log (LOG_WARNING, "Max number of channels reached\n");
        return -1;
    }
    
    /* Line already configured */
    if (find_line (lineID) != NULL) {
        ast_log (LOG_WARNING, "Line %d already configured\n", lineID);
        return -1;
    }


    /* Initialise the interface */
    interfaces[pos].owner[MAINCALL] = NULL;
    interfaces[pos].owner[WAITINGCALL] = NULL;
    interfaces[pos].needhangup = FALSE;
    interfaces[pos].getDigits_thread = AST_PTHREADT_NULL;
    interfaces[pos].timer_thread = AST_PTHREADT_NULL;
    interfaces[pos].callwaiting = callwaiting;
    interfaces[pos].lineID = lineID;
    interfaces[pos].up = FALSE;

    /* Initialising onhook with current time.
       This variable is used to detect flashhook event */
    gettimeofday (&interfaces[pos].onhooktime, NULL);


/* Line IDs 0 to 14 are used to identify FXO/FXS lines. Lines 0, 5 and 10 are FXOs.
 Refer to icp_analogdrv.h for the mapping between line IDs and chips on the voice
 mezzanine cards.

Line IDs for digital lines are defined as 1xyzz where:
- x is the slot number in which the mezzanine is plugged into [0-2]
- y is the DIGITAL line number on the mezzanine card [0-3]
- zz is the timeslot [1-24] for T1 and [1-15,17-31] for E1
*/
    /* DIGITAL */
    if (lineID >= DIGITAL_LINE_MASK) {
        interfaces[pos].mode = DIGITAL;
        nbDigital++;
        lineID -= DIGITAL_LINE_MASK;
        interfaces[pos].slotID = lineID / SLOT_MASK;
        
        lineID %= SLOT_MASK;
        interfaces[pos].spanID = lineID / SPAN_MASK;
                
        interfaces[pos].timeslot = lineID % SPAN_MASK;
        interfaces[pos].echocancel = FALSE;
        
        /* Check lines allocation with port setting */
        if ((hssPorts[interfaces[pos].slotID].setting == SINGLE) &&
            (interfaces[pos].spanID != 0)) {
            ast_log (LOG_ERROR, "Invalid span. Port is set to SINGLE mode\n");
            return -1;
        }
    }
    /* FXO */
    else if (lineID % 5 == 0) {
        interfaces[pos].mode = FXO;
        nbFxsFxo++;
        interfaces[pos].slotID = lineID / 5;
        interfaces[pos].echocancel = echocancel;

        features |= DSP_FEATURE_BUSY_DETECT;
    }
    /* FXS */
    else {
        interfaces[pos].mode = FXS;
        nbFxsFxo++;
        interfaces[pos].slotID = lineID / 5;
        interfaces[pos].echocancel = echocancel;
    }

    /* Trying to define more channels than the limit */
    if (nbFxsFxo > MAX_FXSFXO_NUM || nbDigital > MAX_DIGITAL_NUM) {
        ast_log (LOG_WARNING, "Max number of channels reached\n");
        return -1;
    }
    
    /* Check slot validity */
    if (interfaces[pos].slotID >= NB_HSS_PORTS) {
        ast_log (LOG_ERROR, "Invalid slot %d\n", interfaces[pos].slotID);
        return -1;
    }
    
    /* Check span validity */
    if (interfaces[pos].spanID >= NB_SPANS_PER_MEZZANINE) {
        ast_log (LOG_ERROR, "Invalid span %d\n", interfaces[pos].spanID);
        return -1;
    }
    
    interfaces[pos].format = configFormat;
    
    ast_copy_string (interfaces[pos].cid_num, callerid_num,
                     sizeof (interfaces[pos].cid_num));
                     
    ast_copy_string (interfaces[pos].cid_name, callerid_name,
                     sizeof (interfaces[pos].cid_name));
                     
    ast_copy_string (interfaces[pos].context, context, sizeof (interfaces[pos].context));

    memset (interfaces[pos].ext, 0, sizeof (interfaces[pos].ext));

    /* Init the DSP */
    interfaces[pos].dsp = ast_dsp_new ();

    if (silencesuppression) {
        features |= DSP_FEATURE_SILENCE_SUPPRESS;
    }

    ast_dsp_set_features (interfaces[pos].dsp, features);
    ast_dsp_set_threshold (interfaces[pos].dsp, SILENCE_THRESHOLD);
    ast_dsp_digitmode (interfaces[pos].dsp, DSP_DIGITMODE_DTMF);
    ast_dsp_set_busy_pattern (interfaces[pos].dsp, BUSY_PATTERN, BUSY_PATTERN);
    ast_dsp_set_busy_count (interfaces[pos].dsp, BUSY_COUNT);

    
    if (interfaces[pos].echocancel) {
#ifdef USE_OSLEC_EC
        /* Init the echo canceller */
        interfaces[pos].ec = echo_can_create (TAPS, ECHO_CAN_USE_ADAPTION |
                                                    ECHO_CAN_USE_CLIP | 
                                                    ECHO_CAN_USE_NLP  | 
                                                    ECHO_CAN_USE_TX_HPF);
                                                    
        echo_can_flush (interfaces[pos].ec);

        if (interfaces[pos].ec == NULL) {
            ast_log (LOG_ERROR, "Echo Canceller creation failed\n");
            return -1;
        } else {
            ast_log (LOG_DEBUG, "Echo Canceller creation succeeded\n");
        }
#else
        ast_log (LOG_ERROR, "Trying to use echo cancellation but chan_hss is not linked "
                            "with OSLEC! Recompile with USE_OSLEC_EC flag turned on\n");
        return -1;
#endif
    }

    /* Open a Voice Driver file descriptor */
    interfaces[pos].fd = open (VOICE_DEVICE, O_RDWR);


    if (interfaces[pos].fd <= 0) {
        ast_log (LOG_ERROR, "Can't open Voice driver file descriptor\n");
        return -1;
    }

    if (hssPorts[interfaces[pos].slotID].configured == FALSE) {
        ast_log (LOG_ERROR, "HSS Port %d not configured\n", interfaces[pos].slotID);
        return -1;
    }
    
    /* Bring the HSS port up */
    if (hssPorts[interfaces[pos].slotID].up == FALSE) {
        if (ioctl (interfaces[pos].fd, ICP_HSSVOICEDRV_PORT_UP,
                   &hssPorts[interfaces[pos].slotID].config)) {
            ast_log (LOG_ERROR, "Can't bring up the HSS Port\n");
            return -1;
        }
        
        hssPorts[interfaces[pos].slotID].up = TRUE;
    }


    if (interfaces[pos].mode == DIGITAL) {
        if (setup_digital (&interfaces[pos])) {
            ast_log (LOG_ERROR, "Fail to setup DIGITAL channel\n");
            return -1;
        }
    } else {
        if (setup_pcm (&interfaces[pos])) {
            ast_log (LOG_ERROR, "Fail to setup the HSS phone\n");
            return -1;
        }
    }

    ast_log (LOG_DEBUG, "Creating interface: n %d\n", pos);
    ast_log (LOG_DEBUG, "callerID: %s %s\n", interfaces[pos].cid_name,
             interfaces[pos].cid_num);
    ast_log (LOG_DEBUG, "context: %s\n", interfaces[pos].context);
    ast_log (LOG_DEBUG, "slot: %d   span: %d   line ID: %d\n", interfaces[pos].slotID,
             interfaces[pos].spanID, interfaces[pos].lineID);
    ast_log (LOG_DEBUG, "file descriptor: %d\n\n", interfaces[pos].fd);

    return 0;
}







/***************************************************************/
/*! \brief Parse line IDs values and ranges defined in hss.conf
           and create interfaces
    \param s String to parse
    \return 0 on success else -1 */
/***************************************************************/
static int build_channels (char *s)
{
    char *piece = NULL;
    char *stringToParse = NULL, *baseAddress = NULL;
    int start = 0, finish = 0, x = 0;

    
    if (*s == '\0') {
        ast_log (LOG_ERROR, "Null value for line ID\n");
        return -1;
    }
        
    stringToParse = ast_strdup (s);
    
    if (stringToParse == NULL) {
        ast_log (LOG_ERROR, "ast_strdup failed\n");
        return -1;
    }

    /* save base address of allocated string to free memory */
    baseAddress = stringToParse;
    
    while ((piece = strsep (&stringToParse, ","))) {
        if (sscanf (piece, "%d-%d", &start, &finish) == 2) {
            /* Range */
            if (finish - start >= MAX_TIMESLOT) {
                ast_log (LOG_ERROR, "Error in range definition: too many timeslots\n");
                continue;
            }

        } else if (sscanf (piece, "%d", &start)) {
            /* Just one */
            finish = start;
        } else {
            ast_log (LOG_ERROR,
                     "Syntax error parsing group configuration '%s' at '%s'. Ignoring.\n",
                     s, piece);
            continue;
        }

        for (x = start; x <= finish; x++) {
            ast_log (LOG_DEBUG, "creating line %d\n", x);

            if (mkif (nbInterfaces, x) == -1) {
                ast_log (LOG_ERROR, "Fail to init interface\n");
                free (baseAddress);
                return -1;
            }

            nbInterfaces++;
        }
    }
    
    free (baseAddress);
    return 0;

}


/***************************************************************/
/*! \brief Allocate a new Asterisk channel
    \param pvt Interface to create the channel for
    \param index Either MAINCALL or WAITINGCALL
    \param state State of the Asterisk channel
    \param context Context associated to the channel
    \return The allocated channel or NULL if failed */
/***************************************************************/
static struct ast_channel *newChannel (struct hss_pvt *pvt, call_type index, int state,
                                       char *context)
{
    struct ast_channel *chan = NULL;

    /* Call ast_channel_alloc() to allocate a new Asterisk channel */
    chan = ast_channel_alloc (1, state, pvt->cid_num, pvt->cid_name, "", pvt->ext,
                              context, 0, "HSS/%d-%d", pvt->lineID, index);

    ast_log (LOG_DEBUG, "Creating new HSS channel :%p\n", chan);

    if (chan != NULL) {

        /* Initialise channel structure */
        chan->tech = &hss_tech;

        if (index == MAINCALL)
            chan->fds[0] = pvt->fd;
        else
            chan->fds[0] = -1;

        chan->nativeformats = pvt->format;
        chan->rawreadformat = pvt->format;
        chan->rawwriteformat = pvt->format;
        chan->readformat = pvt->format;
        chan->writeformat = pvt->format;


        /* Don't use ast_set_callerid() here because it will generate a NewCallerID event
           before the NewChannel event */
        chan->cid.cid_num = ast_strdup (pvt->cid_num);
        chan->cid.cid_ani = ast_strdup (pvt->cid_num);
        chan->cid.cid_name = ast_strdup (pvt->cid_name);

        /* Make tech_pvt (from the new channel) point to the hss_pvt structure */
        chan->tech_pvt = pvt;

        /* Make owner (from the hss_pvt structure) point to the created channel */
        pvt->owner[index] = chan;

        pvt->needhangup = FALSE;

        /* Initialise buffer indexes for EC */
        /* Set to -1 to notify that we haven't written data to the voice driver yet */
        pvt->indexWrite = -1;    
        pvt->indexRead = 0;

        ast_atomic_fetchadd_int (&usecnt, 1);
        ast_update_use_count ();

        /* Incoming call : start the PBX */
        if (state != AST_STATE_DOWN) {
            /* Call ast_pbx_start() on this channel to execute the dialplan */
            if (ast_pbx_start (chan)) {
                ast_log (LOG_ERROR, "Unable to start PBX on %s\n", chan->name);
                ast_hangup (chan);
            }
        }
    } else {
        ast_log (LOG_ERROR, "Unable to allocate channel\n");
    }

    return chan;
}


/***************************************************************/
/*! \brief Read data from the voice driver
    \param pvt Interface to read from
    \return Pointer to the read frame else NULL */
/***************************************************************/
inline static struct ast_frame *__hss_read (struct hss_pvt *pvt)
{
    int bytesRead = 0;
    int i = 0;
    struct ast_frame *rxFrame = NULL;
    int localfd = pvt->fd;
    int localformat = pvt->format;
    unsigned char localrxBuffer[VOICE_LINEAR_PACKET_SIZE] = "\0";

#ifdef USE_OSLEC_EC
    short rxSample = 0;
    short txOutFilter = 0;
    short sOutCanceller = 0;
    short txSample = 0;
    unsigned char localtxBuffer[(VOICE_PAYLOAD_SIZE * 2)] = "\0";
#endif
    
    ast_mutex_lock (&pvt->lock);

    /* User has hung up */
    if (pvt->needhangup) {
        ast_log (LOG_DEBUG, "User has hung up\n");
        pvt->needhangup = FALSE;
        ast_mutex_unlock (&pvt->lock);
        return NULL;
    }
    ast_mutex_unlock (&pvt->lock);

    /* Blocking read on the file descriptor to read voice data */
    if (localformat == AST_FORMAT_SLINEAR) {
        bytesRead = read (localfd, localrxBuffer, VOICE_LINEAR_PACKET_SIZE);

        for (i = VOICE_HEADER_SIZE; i < bytesRead; i += 2) {
            SWAP_BYTES (localrxBuffer[i], localrxBuffer[i + 1]);
        }
    } else {
        bytesRead = read (localfd, localrxBuffer, VOICE_ULAW_PACKET_SIZE);
    }

    if (bytesRead < 0) {
        ast_log (LOG_ERROR, "Read error!\n");
        return NULL;
    } else if (bytesRead == 0) {
        ast_log (LOG_ERROR, "Read returned 0 in blocking mode!\n");
        return NULL;
    }
    

    pvt->inFrame.frametype = AST_FRAME_VOICE;
    pvt->inFrame.subclass = pvt->format;
    pvt->inFrame.src = "HSS";
    /* We always read 80 samples, whatever their length */
    pvt->inFrame.samples = VOICE_PAYLOAD_SIZE;

    pvt->inFrame.data = &(pvt->rxBuffer[VOICE_HEADER_SIZE]);

    pvt->inFrame.datalen = (bytesRead - VOICE_HEADER_SIZE);
    pvt->inFrame.offset = 0;
    pvt->inFrame.mallocd = 0;
    pvt->inFrame.delivery = ast_tv (0, 0);

#ifdef USE_OSLEC_EC
    /* Perform Echo Cancellation only when the call is up */
    if (pvt->owner[MAINCALL] != NULL) {
        /* Check if EC is enabled and if there is data in the Tx buffer to perform EC */
        if (pvt->echocancel && pvt->indexWrite != -1) {
            
            /* Perform echo cancellation depending on the codec format
               The algorithm is the same: 
               - retrieve Tx and Rx samples in linear format
               - process to high pass filter the Tx sample
               - process a sample through the echo canceller
               - update the data buffer with the output of the echo canceller */
            if (localformat == AST_FORMAT_ULAW) {

                memcpy (localtxBuffer, pvt->txBuffer[pvt->indexRead], VOICE_PAYLOAD_SIZE);

                for (i = 0; i < VOICE_PAYLOAD_SIZE; i++) {

                    txSample = AST_MULAW ((int) localtxBuffer[i]);
                    rxSample = AST_MULAW ((int) localrxBuffer[i + VOICE_HEADER_SIZE]);

                    txOutFilter = echo_can_hpf_tx (pvt->ec, txSample);
                    sOutCanceller = echo_can_update (pvt->ec, txOutFilter, rxSample);

                    localrxBuffer[i + VOICE_HEADER_SIZE] = AST_LIN2MU (sOutCanceller);
                }
                
                /* Update the read index for the next echo cancellation */
                pvt->indexRead++;

            } else if (localformat == AST_FORMAT_ALAW) {
                memcpy (localtxBuffer, pvt->txBuffer[pvt->indexRead], VOICE_PAYLOAD_SIZE);

                for (i = 0; i < VOICE_PAYLOAD_SIZE; i++) {
                    txSample = AST_ALAW ((int) localtxBuffer[i]);
                    rxSample = AST_ALAW ((int) localrxBuffer[i + VOICE_HEADER_SIZE]);

                    txOutFilter = echo_can_hpf_tx (pvt->ec, txSample);
                    sOutCanceller = echo_can_update (pvt->ec, txOutFilter, rxSample);

                    localrxBuffer[i + VOICE_HEADER_SIZE] = AST_LIN2A (sOutCanceller);
                }
                
                /* Update the read index for the next echo cancellation */
                pvt->indexRead++;

            }
            else if (localformat == AST_FORMAT_SLINEAR) {
                memcpy (localtxBuffer, pvt->txBuffer[pvt->indexRead],
                        (VOICE_PAYLOAD_SIZE * 2));

                 /* samples are Little Endian (read in Big Endian from voice driver, then swapped) */       
                for (i = 0; i < (VOICE_PAYLOAD_SIZE * 2); i += 2) {
                    txSample = localtxBuffer[i] | (localtxBuffer[i + 1] << SIZE_OF_BYTE);
                    rxSample = localrxBuffer[i + VOICE_HEADER_SIZE] | 
                               (localrxBuffer[i + VOICE_HEADER_SIZE + 1] << SIZE_OF_BYTE);

                    txOutFilter = echo_can_hpf_tx (pvt->ec, txSample);
                    sOutCanceller = echo_can_update (pvt->ec, txOutFilter, rxSample);
                    
                    localrxBuffer[i + VOICE_HEADER_SIZE] = (sOutCanceller & 0x00FF);
                    localrxBuffer[i + VOICE_HEADER_SIZE + 1] = (sOutCanceller & 0xFF00) >> SIZE_OF_BYTE;
                }
                
                /* Update the read index for the next echo cancellation */
                pvt->indexRead += 2;

            }

            pvt->indexRead &= (NB_EC_BUFFERS - 1);
        }
    }
#endif

    /* Copying voice samples from local buffer to pvt->rxBuffer */
    if (localformat == AST_FORMAT_SLINEAR) {
        memcpy (pvt->rxBuffer, localrxBuffer, VOICE_LINEAR_PACKET_SIZE);
    } else if ((localformat == AST_FORMAT_ULAW) || (localformat == AST_FORMAT_ALAW)) {
        memcpy (pvt->rxBuffer, localrxBuffer, VOICE_ULAW_PACKET_SIZE);
    }

    rxFrame = ast_dsp_process (pvt->owner[MAINCALL], pvt->dsp, &pvt->inFrame);

    return rxFrame;
}

/***************************************************************/
/*! \brief Configure an HSS port for analog or digital mezzanine
    \param portID HSS port to configure
    \param mode Type of mezzanine, either SLIC or FRAMER
    \param preconfig Predefined configuration (Framer only)
    \param geoconfig Geo configuration (SLIC only)
    \return 0 on success else -1 */
/***************************************************************/
int configure_port (unsigned int portID, port_mode mode, port_setting setting, 
                    unsigned int preconfig, unsigned int geoconfig)
{

    icp_analogdrv_init_t init;
    icp_status_t ret = ICP_STATUS_FAIL;

    if (portID >= NB_HSS_PORTS) {
        ast_log (LOG_ERROR, "Invalid port number\n");
        return -1;
    }

    if (hssPorts[portID].configured == TRUE) {
        ast_log (LOG_ERROR, "Port already configured\n");
        return -1;
    }

    hssPorts[portID].mode = mode;

    if (mode == SLIC) {
        init.slot = portID;
        init.geo_config = geoconfig;
        init.clk_speed = CLK_SPEED_2MHZ;
      
        hssPorts[portID].config.port_config =
            ICP_HSSDRV_PORT_ANALOG_VOICE_MEZZANINE_CONFIG;

        /* First time we try to configure the SLIC/CODEC driver */
        if (slicfd < 0) {
            /* Open SLIC/CODEC driver file descriptor */
            slicfd = open (SLIC_DEVICE, O_RDONLY);

            if (slicfd < 0) {
                ast_log (LOG_ERROR, "Fail to open SLIC/CODEC driver\n");
                return -1;
            }
        }


        if (ioctl (slicfd, ICP_ANALOGDRV_INIT, &init)) {
            ast_log (LOG_ERROR, "SLIC initialisation failed\n");
            return -1;
        }

        hssPorts[portID].configured = TRUE;

    } else if ((mode == E1) || (mode == T1)) {
    
        if ((setting == SINGLE) || (setting == QUAD)) {
            
            hssPorts[portID].setting = setting;
            
        } else {
            
            ast_log (LOG_ERROR, "Port setting is not set\n");
        }
    
    
        hssPorts[portID].config.port_config =
            ICP_HSSDRV_PORT_HMVIP_FRAMER_MEZZANINE_CONFIG;

        ret = icp_FramerDrvInit (portID);

        if (ret != ICP_STATUS_SUCCESS) {
            ast_log (LOG_ERROR, "Framer initialisation failed\n");
            return -1;
        }

        ret = icp_FramerDrvConfigSet (portID, preconfig);

        if (ret != ICP_STATUS_SUCCESS) {
            ast_log (LOG_ERROR, "Framer configuration failed\n");
            return -1;
        }

        hssPorts[portID].configured = TRUE;
        
    } else {
    
        ast_log (LOG_ERROR, "Port mode is not set\n");
    }

    return 0;
}



/***************************************************************/
/*! \brief Asterisk callback. 
           Start playing a digit on a channel
    \return 0 on success else -1 */
/***************************************************************/
static int hss_digit_begin (struct ast_channel *chan, char digit)
{
    struct hss_pvt *pvt = chan->tech_pvt;
    icp_analogdrv_tone_t toneToPlay;

    if (pvt->mode == FXS)
    {
        ast_log (LOG_DEBUG, "Send %c begin\n", digit);

        toneToPlay.lineId = pvt->lineID;
        toneToPlay.tone_dir = TONE_REVERSE_DIRECTION;
     
        switch (digit) {
        case '0':
        case '1':
        case '2':
        case '3':
        case '4':
        case '5':
        case '6':
        case '7':
        case '8':
        case '9':
            toneToPlay.tone = ICP_ANALOG_DTMF_0 + atoi (&digit);
            break;

        case '#':
            toneToPlay.tone = ICP_ANALOG_DTMF_HASH;
            break;

        case '*':
            toneToPlay.tone = ICP_ANALOG_DTMF_STAR;
            break;

        default:
            ast_log (LOG_ERROR, "Trying to play an invalid digit: %c\n", digit);
            return -1;
        }

        /* Start playing the DTMF tone */
        if (ioctl (slicfd, ICP_ANALOGDRV_TONE_ON, &toneToPlay)) {
            ast_log (LOG_ERROR, "Fail to play tone\n");
            return -1;
        }

        return 0;
        
    } else {
        ast_log (LOG_DEBUG, "Can't play the digit. Let Asterisk do it\n");
        return -1;

    }


}



/***************************************************************/
/*! \brief Asterisk callback. 
           Stop playing a digit on a channel
    \return 0 on success else -1 */
/***************************************************************/
static int hss_digit_end (struct ast_channel *chan, char digit, unsigned int duration)
{
    struct hss_pvt *pvt = chan->tech_pvt;

    if (pvt->mode == FXS) {
        ast_log (LOG_DEBUG, "Send %c end\n", digit);

        /* Stop playing the DTMF tone */
        if (ioctl (slicfd, ICP_ANALOGDRV_TONE_OFF, pvt->lineID)) {
            ast_log (LOG_ERROR, "Fail to stop playing tone\n");
            return -1;
        }

        return 0;
    } else {
        ast_log (LOG_DEBUG, "Can't stop playing the digit. Let Asterisk do it\n");
        return -1;
    }
}



/***************************************************************/
/*! \brief Asterisk callback. 
           Indicate a special condition
    \return 0 if condition is handled else -1 (Asterisk handles it) */
/***************************************************************/
static int hss_indicate (struct ast_channel *chan, int condition, const void *data,
                         size_t datalen)
{

    int res = 0;
    struct hss_pvt *pvt = chan->tech_pvt;
    icp_analogdrv_tone_t toneToPlay;

    toneToPlay.tone_dir = TONE_REVERSE_DIRECTION;

    switch (condition) {
    case AST_CONTROL_RINGING:
        if (pvt->mode == FXS) {
           
            /* Play ringback tone */
            toneToPlay.lineId = pvt->lineID;
            toneToPlay.tone = ICP_ANALOG_DTMF_RINGBACK;

            if (ioctl (slicfd, ICP_ANALOGDRV_TONE_ON, &toneToPlay)) {
                ast_log (LOG_ERROR, "Fail to play tone\n");
                res = -1;
            }
        }
        else if (pvt->mode == DIGITAL) {
        
        /* Bring the voice channel up to play the ringback tone to the other end */
        BRING_CHANNEL_UP(pvt);
        res = -1;
        }
        /* The SLIC/CODEC driver can't play tones over FXO.
           Returning -1 so asterisk handles it */
        else
            res = -1;
        break;
    
    case AST_CONTROL_CONGESTION:
    case AST_CONTROL_BUSY:
        if (pvt->mode == FXS) {
            /* Play busy tone */
            toneToPlay.lineId = pvt->lineID;
            toneToPlay.tone = ICP_ANALOG_DTMF_BUSY;

            if (ioctl (slicfd, ICP_ANALOGDRV_TONE_ON, &toneToPlay)) {
                ast_log (LOG_ERROR, "Fail to play tone\n");
                res = -1;
            }

            /* Hangup the channel */
            ast_softhangup (chan, AST_SOFTHANGUP_DEV);
        }
        else if (pvt->mode == DIGITAL) {
            /* Bring the voice channel up to send DTMF digits to the other end */
            BRING_CHANNEL_UP(pvt);
            res = -1;
        } else
            res = -1;
        break;

    case AST_CONTROL_HOLD:
        ast_moh_start (chan, data, "default");
        break;

    case AST_CONTROL_UNHOLD:
        ast_moh_stop (chan);
        break;

    case -1:
        if (pvt->mode == FXS) {
            /* Stop playing tone */
            if (ioctl (slicfd, ICP_ANALOGDRV_TONE_OFF, pvt->lineID)) {
                ast_log (LOG_ERROR, "Fail to stop playing tone\n");
                res = -1;
            }
        }
        break;

    default:
        ast_log (LOG_WARNING, "This indication %d is not supported\n", condition);
        res = -1;
    }

    return res;
}


/***************************************************************/
/*! \brief Asterisk callback. 
           Create an outgoing call to an HSS interface
    \return 0 on success else -1 */
/***************************************************************/
static int hss_call (struct ast_channel *chan, char *dest, int timeout)
{

    struct hss_pvt *pvt = chan->tech_pvt;
    char *digit = NULL;
    icp_analogdrv_tone_t toneToPlay;
    icp_status_t status = ICP_STATUS_FAIL;

    ast_log (LOG_DEBUG, "Outgoing call on line %d from %s:%s\n", pvt->lineID,
             chan->cid.cid_name, chan->cid.cid_num);

    /* Check that the line is down */
    if (chan->_state != AST_STATE_DOWN) {
        ast_log (LOG_WARNING, "Line is not down\n");
      
        return -1;
    }

    ast_mutex_lock (&iflock);

    if (pvt->mode == FXS) {
        /* It is the main call, ring the phone */
        if (pvt->owner[WAITINGCALL] == NULL) {
           
            if (ioctl (slicfd, ICP_ANALOGDRV_RING_ON, pvt->lineID)) {
                ast_log (LOG_ERROR, "Fail to ring line %d\n", pvt->lineID);
                ast_mutex_unlock (&iflock);
                return -1;
            }
        }
        /* We have a waiting call, play the callwaiting tone */
        else {
            toneToPlay.tone_dir = TONE_REVERSE_DIRECTION;
            toneToPlay.lineId = pvt->lineID;
            toneToPlay.tone = ICP_ANALOG_DTMF_CALL_WAITING;

            if (ioctl (slicfd, ICP_ANALOGDRV_TONE_ON, &toneToPlay)) {
                ast_log (LOG_ERROR, "Fail to play tone\n");
                ast_mutex_unlock (&iflock);
                return -1;
            }
        }

        /* Set the state to AST_STATE_RINGING */
        ast_setstate (chan, AST_STATE_RINGING);
        ast_queue_control (chan, AST_CONTROL_RINGING);
        
    } else if (pvt->mode == FXO) {
        
        /* Retrieve the extension to dial */        
        digit = strchr (dest, '/');
        
        if (digit == NULL) {
            ast_log(LOG_WARNING, "No extension to call over FXO line %d\n", pvt->lineID);
            ast_mutex_unlock (&iflock);
            return -1;
        }
        
        digit++;

        /* Send OFFHOOK command */
        if (ioctl (slicfd, ICP_ANALOGDRV_OFF_HOOK, pvt->lineID)) {
            ast_log (LOG_ERROR, "Fail to send OFFHOOK on FXO\n");
            ast_mutex_unlock (&iflock);
            return -1;
        }
         
         /* FXO answers the call, set the channel to AST_STATE_UP */   
        ast_setstate (chan, AST_STATE_UP);
        ast_queue_control (chan, AST_CONTROL_ANSWER);
        
        /* wait for the other end to answer */
        usleep(1000000);
          
        BRING_CHANNEL_UP(pvt);
        
        if (pvt->up != TRUE) {
            ast_mutex_unlock (&iflock);
            return -1;
        }
        
        ast_log (LOG_DEBUG, "Sending extension %s over line %d\n", digit, pvt->lineID);

        /* Send the digits */
        while (*digit != '\0') {
            ast_senddigit (chan, *digit);
            digit++;
            ast_safe_sleep(chan, INTERDIGITTIME);
        }
        
    } else if (pvt->mode == DIGITAL) {

        /* Retrieve the extension to dial */       
        digit = strchr (dest, '/');
        
        if (digit == NULL) {
            ast_log(LOG_WARNING, "No extension to call over digital line %d\n", 
                    pvt->lineID);
            ast_mutex_unlock (&iflock);
            return -1;
        }
        
        digit++;
        
        /* Send OFFHOOK signal over the digital line */
        status = icp_FramerDrvSignalStateSet (pvt->slotID, pvt->spanID, pvt->timeslot,
                                              EM_OFFHOOK);

        if (status != ICP_STATUS_SUCCESS) {
            ast_log (LOG_ERROR,
                     "Fail to change signal state on slot %d, span %d, ts %d\n",
                     pvt->slotID, pvt->spanID, pvt->timeslot);
            ast_mutex_unlock (&iflock);
            return -1;
        }

        /* wait for the other end to answer */
        usleep(200000);
        
        /* Bring the voice channel up to send DTMF digits to the other end */
        BRING_CHANNEL_UP(pvt);
        
        if (pvt->up != TRUE) {
            ast_mutex_unlock (&iflock);
            return -1;
        }

        /* Send the digits */
        while (*digit != '\0') {
            ast_senddigit (chan, *digit);
            digit++;
            ast_safe_sleep(chan, INTERDIGITTIME);
        }

        /* The other end is now ringing, set the channel state to AST_STATE_RINGING */
        ast_setstate (chan, AST_STATE_RINGING);
    }

    ast_mutex_unlock (&iflock);

    return 0;
}



    
/***************************************************************/
/*! \brief Asterisk callback. 
           Notify that the callee has answered the call
    \return 0 on success else -1 */
/***************************************************************/
static int hss_answer (struct ast_channel *chan)
{
    struct hss_pvt *pvt = chan->tech_pvt;
    icp_status_t status = ICP_STATUS_FAIL;
    
    if (pvt->mode == FXO) {
        if (ioctl (slicfd, ICP_ANALOGDRV_OFF_HOOK, pvt->lineID) != 0) {
            ast_log (LOG_ERROR, "Failed to send OFF HOOK over FXO %d\n", pvt->lineID);
            return -1;
        }

        ast_log (LOG_DEBUG, "Sending off hook\n");
        
    } else if (pvt->mode == DIGITAL) {
        status =
            icp_FramerDrvSignalStateSet (pvt->slotID, pvt->spanID, pvt->timeslot,
                                         EM_OFFHOOK);

        if (status != ICP_STATUS_SUCCESS) {
            ast_log (LOG_ERROR,
                     "icp_FramerDrvSignalSet failed on slot %d, span %d, timeslot %d\n",
                     pvt->slotID, pvt->spanID, pvt->timeslot);
            return -1;
        }
        
        ast_log (LOG_DEBUG, "Sending OFFHOOK on DIGITAL link\n");
    }
    
    BRING_CHANNEL_UP(pvt);
    
    ast_setstate (chan, AST_STATE_UP);

    return 0;
}




/***************************************************************/
/*! \brief Asterisk callback. 
           Write data to the Voice driver
    \return 0 on success else -1 */
/***************************************************************/
static int hss_write (struct ast_channel *chan, struct ast_frame *frame)
{

    struct hss_pvt *pvt = chan->tech_pvt;
    struct ast_frame localframe = *frame;
    int i;
    
#ifdef USE_OSLEC_EC
    int localindexWrite = 0;
    unsigned int availableSpace = 0;
#endif
    
    int bytesWritten = 0;
    unsigned char buf[(VOICE_PAYLOAD_SIZE * NB_EC_BUFFERS)];

    /* return 0 if the voice channel is disabled */
    if (pvt->up == FALSE) {
        return 0;
    }
    
    if (localframe.frametype != AST_FRAME_VOICE) {
        ast_log (LOG_WARNING, "Can't handle this frametype\n");
        return 0;
    }

    if (!localframe.data || !localframe.datalen) {
        ast_log (LOG_DEBUG, "Frame doesn't contain data, there is nothing to write\n");
        return 0;
    }

    if (localframe.datalen > (sizeof(buf) - VOICE_HEADER_SIZE)) {
        ast_log (LOG_WARNING, "Frame length greater than buffer space\n");
        return 0;
    }

    /* If Asterisk tries to write on the waiting channel, discard the frame as the
       waiting channel is on hold */
    if (pvt->owner[MAINCALL] != chan) {
        return 0;
    }

    /* Add headers for voice data */
    /* Data length */
    buf[0] = (localframe.datalen & 0xFF00) >> SIZE_OF_BYTE;
    buf[1] = localframe.datalen & 0x00FF;
    
    /* Channel ID */
    buf[2] = 0;
    buf[3] = 0;

    memcpy (&buf[VOICE_HEADER_SIZE], localframe.data, localframe.datalen);


    if (localframe.subclass == AST_FORMAT_SLINEAR) {
        for (i = VOICE_HEADER_SIZE; i < localframe.datalen + VOICE_HEADER_SIZE; i += 2) {
            SWAP_BYTES (buf[i], buf[i + 1]);
        }
    }

    bytesWritten = write (pvt->fd, buf, (localframe.datalen + VOICE_HEADER_SIZE));

#ifdef USE_OSLEC_EC
    /* Echo cancellation: fill the buffer with written data */
    if (pvt->indexWrite == -1)
        pvt->indexWrite = 0;
   
    localindexWrite = pvt->indexWrite;
 
    /* compute the contiguous space left in the circular buffer to reduce the number of 
       calls to memcpy() */
    availableSpace = (NB_EC_BUFFERS - localindexWrite) * VOICE_PAYLOAD_SIZE;
    
    /* if there is enough space, write all the voice samples at once */
    if (availableSpace >= localframe.datalen) {
    
        memcpy (&pvt->txBuffer[localindexWrite], localframe.data, localframe.datalen);
        localindexWrite += (localframe.datalen / VOICE_PAYLOAD_SIZE);

    /* there is not enough space so we call memcpy() twice */
    } else {
        /* write until we reach the end of the buffer */
        memcpy (&pvt->txBuffer[localindexWrite], localframe.data, availableSpace);
        localindexWrite = 0;
        
        /*write the rest */
        memcpy (&pvt->txBuffer[localindexWrite], (localframe.data + availableSpace), (localframe.datalen - availableSpace));
        localindexWrite += ((localframe.datalen - availableSpace) / VOICE_PAYLOAD_SIZE);
    }
    
    localindexWrite &= (NB_EC_BUFFERS - 1);

    pvt->indexWrite = localindexWrite;
#endif

    return 0;
}




/***************************************************************/
/*! \brief Asterisk callback. 
           Read data from the Voice driver
    \return Read frame else NULL */
/***************************************************************/
static struct ast_frame *hss_read (struct ast_channel *chan)
{

    struct hss_pvt *pvt = chan->tech_pvt;
    struct ast_frame *f = NULL;

    /* Read frames only for the main channel and only when the call is established */
    if (pvt->owner[MAINCALL] == chan) {
      
        /* Calling the inline function to read data from the voice driver */
        f = __hss_read (pvt);
    }
 
    /* If this is not the main channel it means we're trying to read for the
       callwaiting channel, so we don't read as asterisk is playing MOH. */
    else {
        ast_mutex_lock (&pvt->lock);

        if (pvt->needhangup) {
      
            ast_log (LOG_DEBUG, "User has hung up line %d\n", pvt->lineID);
            pvt->needhangup = FALSE;
          
        } else {

/* We are trying to read data for the waiting call (which is on hold), so return
   an empty frame */
            f = &ast_null_frame;
        }

        ast_mutex_unlock (&pvt->lock);
    }


    return f;
}




/***************************************************************/
/*! \brief Asterisk callback. 
           Hang up a channel and check for waiting call
    \return 0 on success else -1 */
/***************************************************************/
static int hss_hangup (struct ast_channel *chan)
{

    struct hss_pvt *pvt = NULL;
    icp_analogdrv_tone_t toneToPlay;
    int hookstate = FXS_ONHOOK;
    icp_status_t status = ICP_STATUS_FAIL;

    
    if (chan == NULL) {
        ast_log(LOG_NOTICE, "Channel already hung up, nothing to do\n");
        return 0;
    }    
    
    pvt = chan->tech_pvt;
    
    ast_mutex_lock (&pvt->lock);
    
    ast_dsp_reset (pvt->dsp);

    /* Set the owner field of pvt struct to NULL */
    if (pvt->owner[MAINCALL] == chan) {
        pvt->owner[MAINCALL] = NULL;
    } else {
        pvt->owner[WAITINGCALL] = NULL;
    }

    ast_mutex_unlock (&pvt->lock);
    
/* Hanging up the call, so breaking the link between the asterisk channel
   and the private structure */
    chan->tech_pvt = NULL;

/* Update the count of channels in use */
    ast_atomic_fetchadd_int (&usecnt, -1);
    ast_update_use_count ();

    /* Reset extension number */
    bzero (pvt->ext, sizeof (pvt->ext));

    /*Set state to AST_STATE_DOWN */
    ast_setstate (chan, AST_STATE_DOWN);

    if (pvt->mode == FXS) {

        hookstate = ioctl (slicfd, ICP_ANALOGDRV_HOOKSTATE_GET, pvt->lineID);
        
        /* Stop ringing in case the caller hangs up before the callee picks up */
        if (hookstate == FXS_ONHOOK) {
            if (ioctl (slicfd, ICP_ANALOGDRV_RING_OFF, pvt->lineID)) {
                ast_log (LOG_ERROR, "Fail to ring line %d\n", pvt->lineID);
                return -1;
            }
        }

        /* If phone offhook and no main call, play busy signal */
        if ((hookstate == FXS_OFFHOOK) && (pvt->owner[MAINCALL] == NULL)) {
            ast_log (LOG_DEBUG, "Call is over. Playing busy tone\n");

            toneToPlay.lineId = pvt->lineID;
            toneToPlay.tone = ICP_ANALOG_DTMF_BUSY;
            toneToPlay.tone_dir = TONE_REVERSE_DIRECTION;

            if (ioctl (slicfd, ICP_ANALOGDRV_TONE_ON, &toneToPlay)) {
                ast_log (LOG_ERROR, "Fail to play tone\n");
                return -1;
            }
            
            BRING_CHANNEL_DOWN(pvt);
            
            if (pvt->up != FALSE) {
                return -1;
            }            
                        
        }
        /* If there is a waiting call and the phone is onhook, ring the line */
        else if (pvt->owner[WAITINGCALL] != NULL) {
            ast_log (LOG_DEBUG, "Waiting call! ringing the line again\n");

            if (ioctl (slicfd, ICP_ANALOGDRV_RING_ON, pvt->lineID)) {
                ast_log (LOG_ERROR, "Fail to ring line %d\n", pvt->lineID);
                return -1;
            }

            switch_channels (pvt);

        }
        /* All calls terminated, bring down the voice channel */
        else if ((pvt->owner[MAINCALL] == NULL) && (pvt->owner[WAITINGCALL] == NULL)) {
            ast_log (LOG_DEBUG, "All calls over\n");

            BRING_CHANNEL_DOWN(pvt);
            
            if (pvt->up != FALSE) {
                return -1;
            }            
            
        } else {

            if (ioctl (slicfd, ICP_ANALOGDRV_TONE_OFF, pvt->lineID)) {
                ast_log (LOG_ERROR, "Fail to stop playing tone\n");
                return -1;
            }
        }
        
    } else if (pvt->mode == DIGITAL) {
        ast_log (LOG_DEBUG, "Sending ONHOOK on DIGITAL link\n");

        status =
            icp_FramerDrvSignalStateSet (pvt->slotID, pvt->spanID, pvt->timeslot,
                                         EM_ONHOOK);

        if (status != ICP_STATUS_SUCCESS) {
            ast_log (LOG_ERROR, "Fail to set signal on slot %d, span %d, timeslot %d\n",
                     pvt->slotID, pvt->spanID, pvt->timeslot);
            return -1;
        }

        BRING_CHANNEL_DOWN(pvt);
        
        if (pvt->up != FALSE) {
                return -1;
        }


    } else if (pvt->mode == FXO) {
        BRING_CHANNEL_DOWN(pvt);
        
        if (pvt->up != FALSE) {
                return -1;
        }
        
        ast_log (LOG_DEBUG, "Sending ONHOOK on FXO\n");

        if (ioctl (slicfd, ICP_ANALOGDRV_ON_HOOK, pvt->lineID)) {
            ast_log (LOG_ERROR, "Fail to send ONHOOK event on line %d\n", pvt->lineID);
            return -1;
        }

    }

    return 0;
}




/***************************************************************/
/*! \brief Asterisk callback. 
           Request to allocate a channel for a call
    \return The new asterisk channel else NULL if failed */
/***************************************************************/
static struct ast_channel *hss_request (const char *type, int format, void *data,
                                        int *cause)
{
    int requestedLine = -1;
    struct hss_pvt *pvt = NULL;
    struct ast_channel *chan = NULL;
    unsigned int slotID = 0, spanID = 0;
    char *dest = NULL;
    uint8_t state = 0;
    icp_status_t status = ICP_STATUS_FAIL;
    int nbValues = 0;

    dest = ast_strdup ((char *) data);
    
    if (dest == NULL) {
        ast_log (LOG_ERROR, "ast_strdup failed\n");
        return NULL;
    }
    
    ast_log (LOG_DEBUG, "Requesting line %s\n", dest);

    ast_mutex_lock (&iflock);

/*There are 2 possibilities to make a call from Asterisk:
 - make a call over a span (DIGITAL lines only), ie use the first available timeslot on this span
 - make a call over a line, ie using a specific line. This can be an FXS/FXO line or a 
 digital line (refer to comments in mkif() function to see how digital lines are defined)  */


    /* Call over a span: find the first available timeslot */
    if (dest[0] == 's') {
       
        nbValues = sscanf (&dest[1], "%1d%1d", &slotID, &spanID);

        if (nbValues != 2) {
            ast_log (LOG_ERROR, "Invalid digital line ID\n");
            ast_mutex_unlock (&iflock);
            free (dest);
            return NULL;
        }

        pvt = find_available_line_on_span (slotID, spanID);

        if (pvt == NULL) {
            ast_log (LOG_ERROR, "No line available on slot %d, span %d\n", slotID,
                     spanID);
            *cause = AST_CAUSE_BUSY;
            ast_mutex_unlock (&iflock);
            free (dest);
            return NULL;
        }

        chan = newChannel (pvt, MAINCALL, AST_STATE_DOWN, pvt->context);

    }
    /* Call over a line */
    else {
        /*Get the lineID of the channel to search for */
        requestedLine = atoi (dest);

        /* Search for this channel in the array of interfaces */
        pvt = find_line (requestedLine);

        if (pvt == NULL) {
            ast_log (LOG_WARNING, "This line is not configured\n");
            *cause = AST_CAUSE_NO_ROUTE_DESTINATION;
            ast_mutex_unlock (&iflock);
            free (dest);
            return NULL;
        }

        /* This channel has no current call (unowned channel) */
        if (pvt->owner[MAINCALL] == NULL) {
            if (pvt->mode == FXS) {
            
                /* The phone is on hook */
                if (ioctl (slicfd, ICP_ANALOGDRV_HOOKSTATE_GET, pvt->lineID) ==
                    FXS_ONHOOK) {
                    /* Create an Asterisk channel */
                    chan = newChannel (pvt, MAINCALL, AST_STATE_DOWN, pvt->context);

                    if (chan == NULL)
                        ast_log (LOG_ERROR, "Creation failed\n");
                } else
                    *cause = AST_CAUSE_BUSY;
                    
            } else if (pvt->mode == DIGITAL) {
                status = icp_FramerDrvSignalStateGet (pvt->slotID, pvt->spanID,
                                                      pvt->timeslot, &state);

                if (status != ICP_STATUS_SUCCESS) {
                    ast_log (LOG_ERROR,
                             "Fail to retrieve state on slot %d, span %d, timeslot %d\n",
                             pvt->slotID, pvt->spanID, pvt->timeslot);

                }

                if (state == EM_ONHOOK) {
                    /* Create an Asterisk channel */
                    chan = newChannel (pvt, MAINCALL, AST_STATE_DOWN, pvt->context);

                    if (chan == NULL)
                        ast_log (LOG_ERROR, "Creation failed\n");
                }
            } else if (pvt->mode == FXO) {
                /* Create an Asterisk channel */
                if (pvt->owner[MAINCALL] == NULL) {
                    chan = newChannel (pvt, MAINCALL, AST_STATE_DOWN, pvt->context);

                    if (chan == NULL)
                        ast_log (LOG_ERROR, "Creation failed\n");
                        
                } else {
                    ast_log (LOG_ERROR, "Line is busy\n");
                    *cause = AST_CAUSE_BUSY;
                }
            }

        }
        /* FXS only: participating in a call and no current waiting call */
        else if (pvt->callwaiting &&
                 pvt->mode == FXS &&
                 pvt->owner[MAINCALL] != NULL && pvt->owner[WAITINGCALL] == NULL) {
            /* Create a new channel only if the main channel is up
               (avoid creating a channel if the phone is ringing, 
                which call will we answer then?) */
            if (pvt->owner[MAINCALL]->_state == AST_STATE_UP) {
                chan = newChannel (pvt, WAITINGCALL, AST_STATE_DOWN, pvt->context);

                if (chan == NULL)
                    ast_log (LOG_ERROR, "Creation failed\n");
                    
            } else {
                ast_log (LOG_ERROR, "Line is busy\n");
                *cause = AST_CAUSE_BUSY;
            }
            
        } else {
            ast_log (LOG_ERROR, "Line is busy\n");
            /* The channel is currently in use */
            *cause = AST_CAUSE_BUSY;
        }

    }
    
    ast_mutex_unlock (&iflock);
    free (dest);

    return chan;
}




/***************************************************************/
/*! \brief Asterisk callback. 
           Update owner links
    \return 0 on success else -1 */
/***************************************************************/
static int hss_fixup (struct ast_channel *old, struct ast_channel *new)
{

    struct hss_pvt *pvt = old->tech_pvt;

    if (pvt == NULL) {
        ast_log (LOG_ERROR, "Fixup failed. Invalid private structure\n");
        return -1;
    }
    
    ast_mutex_lock (&pvt->lock);

    /*Update the owner link, whether it's the main call or the waiting call */
    if (pvt->owner[MAINCALL] == old) {
        pvt->owner[MAINCALL] = new;
    } else if (pvt->owner[WAITINGCALL] == old) {
        pvt->owner[WAITINGCALL] = new;
    }

    ast_mutex_unlock (&pvt->lock);

    return 0;
}




/***************************************************************/
/*! \brief Handler for the CLI command. 
           Prints out framer driver status.
    \return RESULT_SUCCESS on success, RESULT_SHOWUSAGE for 
            improper arguments, RESULT_FAILURE on failure      */
/***************************************************************/
static int handle_hss_framer_status (int fd, int argc, char *argv[])
{

    icp_status_t status = ICP_STATUS_FAIL;

    if (argc != 3) {
        return RESULT_SHOWUSAGE;
    }

    status = icp_FramerDrvStatus();
    
    if (status != ICP_STATUS_SUCCESS) {
        return RESULT_FAILURE;
    }
    
    return RESULT_SUCCESS;
}

/***************************************************************/
/*! \brief This thread reads voice samples from the voice driver
           and detect digits using Asterisk DSP
    \param data Interface to detect the digits on */
/***************************************************************/
static void *get_digits (void *data)
{

    struct hss_pvt *pvt = data;
    struct pollfd fds;
    struct ast_frame *fr = NULL;
    int retValue = 0;
    int endloop = FALSE;
    int startDial = FALSE;
    int timeout = -1;
    struct timeval lastDigit = {0,0};
    icp_analogdrv_tone_t toneToPlay;
    struct sched_param old;
    int policy;
            
    pthread_getschedparam(pthread_self(), &policy, &old);
    
    ast_log (LOG_DEBUG, "Retrieving digits on line %d...\n", pvt->lineID);
    set_priority (DIGITS_THREAD_PRIORITY);
    
    fds.fd = pvt->fd;
    fds.events = POLLIN;

    /* Reset extension */
    bzero (pvt->ext, sizeof (pvt->ext));
    
    /* Set timeout value based on the type of interface */
    if ((pvt->mode == FXS) || (pvt->mode == FXO)) {
        timeout = ANALOGDIGITTIMEOUT;
    } else {
        timeout = DIGITALDIGITTIMEOUT;
    }
    
    do {
        
        retValue = poll (&fds, 1, POLL_TIMEOUT);

        if (retValue == 0) {
            ast_log (LOG_ERROR, "Polling timeout\n");
            pthread_exit (NULL);
        }
        
        if (retValue < 0) {
            ast_log (LOG_ERROR, "Error in polling\n");
            pthread_exit (NULL);
        }

        fr = __hss_read (pvt);

        if (fr == NULL) {
        pthread_exit (NULL);
        }
        
        /* Digit available */
        if (fr->frametype == AST_FRAME_DTMF) {
        
            /* save the time to detect timeout */
            lastDigit = ast_tvnow();
            
            /* This is the first digit, the user is dialing */
            if (!startDial) {
               
                startDial = TRUE;
                
                 /* First digit on an FXS interface: stop playing the dial tone  */
                if (pvt->mode == FXS) {
                    if (ioctl (slicfd, ICP_ANALOGDRV_TONE_OFF, pvt->lineID)) {
                        ast_log (LOG_ERROR, "Fail to stop playing tone\n");
                        pthread_exit (NULL);
                    }
                }
            }

            ast_log (LOG_DEBUG, "I have a digit! ->%c<-\n", fr->subclass);

            /* Add the digit to the extension */
            sprintf (pvt->ext, "%s%c", pvt->ext, (char) fr->subclass);
            
        }
       
        /* The user has started dialing and last digit was received "timeout" ms ago */
        if ((ast_tvdiff_ms(ast_tvnow(), lastDigit) >= timeout) && startDial) {
        
            /* We're not expecting any more digits, bring down the voice channel */
            BRING_CHANNEL_DOWN(pvt);
            
            pthread_setschedparam(pthread_self(), policy, &old);
            
            /* The extension exists in the associated context */
            if (ast_exists_extension (NULL, pvt->context, pvt->ext, 1, pvt->cid_num)) {
                newChannel (pvt, MAINCALL, AST_STATE_RING, pvt->context);
                endloop = TRUE;
                
            } else {
            
                /* Invalid extension */
                ast_log (LOG_WARNING, "Invalid extension\n");
                endloop = TRUE;

                /* Play busy tone */
                if (pvt->mode == FXS) {
                    toneToPlay.lineId = pvt->lineID;
                    toneToPlay.tone = ICP_ANALOG_DTMF_BUSY;
                    toneToPlay.tone_dir = TONE_REVERSE_DIRECTION;

                    if (ioctl (slicfd, ICP_ANALOGDRV_TONE_ON, &toneToPlay)) {
                        ast_log (LOG_ERROR, "Fail to play tone\n");
                        pthread_exit (NULL);
                    }
                }
                else if (pvt->mode == DIGITAL) {
                    if (ast_exists_extension (NULL, pvt->context, "i", 1, pvt->cid_num)) {
                        ast_log(LOG_WARNING, "Jumping to 'i' extension\n");
                        strcpy(pvt->ext, "i");
                        newChannel (pvt, MAINCALL, AST_STATE_RING, pvt->context);
                        endloop = TRUE;
                    } else {
                        ast_log (LOG_WARNING, "You should add an 'i' extension in your "
                                              "dialplan to handle invalid extensions\n");
                    }
                    
                }
                
            }
        }

    } while (!endloop);
    
    ast_log (LOG_DEBUG, "extension dialed : %s\n", pvt->ext);
    pvt->getDigits_thread = AST_PTHREADT_NULL;

    return NULL;
}



/***************************************************************/
/*! \brief This thread handles ONHOOK event and detects 
           flashhook events
    \param data Interface to detect flashhook on */
/***************************************************************/
static void *handle_hangup (void *data)
{
    struct hss_pvt *pvt = data;
    struct ast_channel *chan = NULL;

    chan = pvt->owner[MAINCALL];

    /* Wait for flashhook time */
    usleep (HOOKFLASHTIME * 1000);

    /* This is a hangup: prevent from external cancellation */
    pthread_setcancelstate (PTHREAD_CANCEL_DISABLE, NULL);

    ast_mutex_lock(&pvt->lock);
    
    if (chan != NULL) {
        /* User hangs the line up (put down the phone) */
        if (chan->_state == AST_STATE_UP || chan->_state == AST_STATE_RING) {
            
            ast_mutex_unlock(&pvt->lock);
    
            ast_softhangup (chan, AST_SOFTHANGUP_DEV);
    
            ast_log (LOG_DEBUG, "Hanging the line up : %d\n", pvt->lineID);
            
        } else {
        
            ast_mutex_unlock(&pvt->lock);
            
            ast_log (LOG_DEBUG, "Line %d already hung up. Releasing mutex\n", pvt->lineID);
        }
        
    } else {
    
        ast_mutex_unlock(&pvt->lock);
    
        ast_log(LOG_DEBUG, "Line %d already hung up\n", pvt->lineID);
    }
       
    pvt->timer_thread = AST_PTHREADT_NULL; 
    
    
    
    return NULL;
}



/*******************************************************************/
/*! \brief This thread monitors the events coming from 
           the SLIC/CODEC driver for all the analog lines */
/*******************************************************************/
static void *do_monitor (void *data)
{

    struct hss_pvt *pvt = NULL;
    pthread_attr_t attr;
    icp_analogdrv_event_t event;
    struct ast_channel *chan = NULL;
    struct timeval now;
    icp_analogdrv_tone_t toneToPlay;
    int ret = 0;
    
    pthread_attr_init (&attr);
    pthread_attr_setdetachstate (&attr, PTHREAD_CREATE_DETACHED);

    ast_log (LOG_DEBUG, "FXS-FXO monitor thread started\n");
    set_priority (MONITOR_THREAD_PRIORITY);
    
    /* Wait for events on the SLIC/CODEC */
    while (TRUE) {

/* Blocking read for events coming from the SLIC/CODEC driver */
        ret = read (slicfd, &event, sizeof (event));

        if (ret < 0) {
            ast_log (LOG_ERROR, "Fail to read event from SLIC/CODEC driver\n");
        }

        if (event.lineId < 0 || event.lineId >= MAX_FXSFXO_NUM) {
            ast_log (LOG_ERROR, "Event on invalid line\n");
            continue;
        }

        pvt = find_line (event.lineId);

        /* If the line is configured */
        if (pvt != NULL) {

            switch (event.eventType) {
            /*******************************************/
            /*             FXS goes OFFHOOK            */
            /*******************************************/
            case ICP_ANALOGDRV_EVENT_OFF_HOOK:

                ast_log (LOG_DEBUG, "Line %d is now off hook\n", event.lineId);
                gettimeofday (&now, NULL);

                /* Do we have a flash hook? */
                if (ast_tvdiff_ms (now, pvt->onhooktime) < HOOKFLASHTIME) {
                    ast_log (LOG_DEBUG, "Got a FLASHHOOK!\n");
                    
                    /* This is a flashhook: stop the timer thread detecting this event */
                    if (pvt->timer_thread != AST_PTHREADT_NULL) {
                        pthread_cancel (pvt->timer_thread);
                        pvt->timer_thread = AST_PTHREADT_NULL;
                    }

                    /* Switch between channels if there is a waiting call */
                    if (pvt->owner[WAITINGCALL] != NULL) {
                        ast_log (LOG_DEBUG, "Someone is waiting\n");

                        if (ioctl (slicfd, ICP_ANALOGDRV_TONE_OFF, pvt->lineID)) {
                            ast_log (LOG_ERROR, "Fail to stop playing tone\n");
                        }

                        /* Put the main call on hold if we have not hanged up */
                        if (pvt->owner[MAINCALL] != NULL) {
                            ast_queue_control (pvt->owner[MAINCALL], AST_CONTROL_HOLD);
                            ast_log (LOG_DEBUG, "%s on hold\n",
                                     pvt->owner[MAINCALL]->name);
                        }

                        ast_queue_control (pvt->owner[WAITINGCALL], AST_CONTROL_UNHOLD);
                        ast_log (LOG_DEBUG, "%s on unhold\n",
                                 pvt->owner[WAITINGCALL]->name);

                        switch_channels (pvt);
                    }

                }
                       
                    /* Bring up the channel */
                    BRING_CHANNEL_UP(pvt);

                /* Receiving scenario: User answers the line */
                /* .owner is used to determine if the interface is on a call or not.
                   _state is used only when a call is in progress. */
                if (pvt->owner[MAINCALL] != NULL) {
                    chan = pvt->owner[MAINCALL];

                    ast_log (LOG_DEBUG, "Answering the line\n");

                    /* Stop ringing */
                    if (ioctl (slicfd, ICP_ANALOGDRV_RING_OFF, pvt->lineID)) {
                        ast_log (LOG_ERROR, "Fail to stop ringing line %d\n",
                                 pvt->lineID);
                    }

                    if (chan->_state == AST_STATE_RINGING) {
                        ast_setstate (chan, AST_STATE_UP);
                        ast_queue_control (chan, AST_CONTROL_ANSWER);
                    } else {
                        ast_queue_control (chan, AST_CONTROL_UNHOLD);
                    }
                }
                /* Sending scenario: User picks up to dial */
                else {

                    /*  Play the dial tone  */
                    toneToPlay.lineId = pvt->lineID;
                    toneToPlay.tone = ICP_ANALOG_DTMF_DIAL;
                    toneToPlay.tone_dir = TONE_REVERSE_DIRECTION;

                    if (ioctl (slicfd, ICP_ANALOGDRV_TONE_ON, &toneToPlay)) {
                        ast_log (LOG_ERROR, "Fail to play tone\n");
                    }

                    /*  Wait for digits */
                    ast_pthread_create (&pvt->getDigits_thread, &attr, get_digits, pvt);
                }
                break;

            /*******************************************/
            /*             FXS goes ONHOOK             */
            /*******************************************/
            case ICP_ANALOGDRV_EVENT_ON_HOOK:

                ast_log (LOG_DEBUG, "Line %d is now on hook\n", event.lineId);

                if (ioctl (slicfd, ICP_ANALOGDRV_TONE_OFF, pvt->lineID)) {
                    ast_log (LOG_ERROR, "Fail to stop playing tone\n");
                }

                /* Get time to determine hookflash */
                gettimeofday (&pvt->onhooktime, NULL);

                /* User hangs up while dialing (no call created) */
                if (pvt->getDigits_thread != AST_PTHREADT_NULL) {
                    /* Kill the thread reading the digits on this structure */
                    pthread_cancel (pvt->getDigits_thread);
                    pvt->getDigits_thread = AST_PTHREADT_NULL;

                    ast_log (LOG_DEBUG, "Thread cancellation\n");

                    /* Bring the voice channel down */
                    BRING_CHANNEL_DOWN(pvt);

                    /* Reset extension number */
                    bzero (pvt->ext, sizeof (pvt->ext));

                }

                /* Caller puts the phone on hook while in a call, detect if it's a hookflash */
                else if (pvt->owner[MAINCALL] != NULL) {

                    ast_pthread_create (&pvt->timer_thread, &attr, handle_hangup, pvt);
                }
                /* Callee has hung up, caller puts the phone down and there is a waiting call */
                else if (pvt->owner[WAITINGCALL] != NULL) {

                    ast_log (LOG_WARNING, "Waiting call! Ringing the line again\n");

                    if (ioctl (slicfd, ICP_ANALOGDRV_RING_ON, pvt->lineID)) {
                        ast_log (LOG_ERROR, "Fail to ring line %d\n", pvt->lineID);
                    }

                    switch_channels (pvt);

                }
                break;

            /*******************************************/
            /*       Incoming call from the CO         */
            /*******************************************/
            case ICP_ANALOGDRV_EVENT_RING_ON:

                ast_log (LOG_DEBUG, "FXO line %d receives ringing\n", event.lineId);

                newChannel (pvt, MAINCALL, AST_STATE_RING, pvt->context);

                break;

            /*******************************************/
            /* CO hangs up before the call is answered */
            /*******************************************/
            case ICP_ANALOGDRV_EVENT_RING_OFF:

                ast_log (LOG_DEBUG, "FXO line %d stops ringing\n", event.lineId);


                if (pvt->owner[MAINCALL] != NULL) {
                    chan = pvt->owner[MAINCALL];
                    
                    /* Lock channel to check _state */
                    ast_channel_lock (chan);

                    /* Hang up the call if the FXO rings a phone */
                    if (chan->_state == AST_STATE_RING) {
                        
                        ast_softhangup_nolock (chan, AST_SOFTHANGUP_DEV);
                        ast_log (LOG_DEBUG, "Hanging up the FXO line\n");
                    }
                    
                    ast_channel_unlock (chan);
                }
                break;

            /*************************************************/
            /* Disconnect Supervision with polarity reversal */
            /*************************************************/
            case ICP_ANALOGDRV_EVENT_POL_REV:
            
                ast_log (LOG_DEBUG, "FXO line %d receives polarity reversal\n", event.lineId);
                
                if (pvt->owner[MAINCALL] != NULL) {
                    
                    chan = pvt->owner[MAINCALL];
                    
                    /* Lock channel to check _state */
                    ast_channel_lock (chan);
       
                    /* Hangup the call */
                    if (chan->_state == AST_STATE_UP) {            
                        
                        ast_softhangup_nolock (chan, AST_SOFTHANGUP_DEV);
                        ast_log (LOG_DEBUG, "Hanging up the FXO line : %d\n", pvt->lineID);
                    }
                    
                    ast_channel_unlock (chan);
                }
                break;

            default:

                ast_log (LOG_NOTICE, "Unidentified event from SLIC/CODEC driver: %d\n",
                         event.eventType);
                break;
            } /* end switch(event.eventType) */
       
        }

    } /* end while(TRUE) */


    return NULL;
}



/*******************************************************************/
/*! \brief This thread monitors the events coming from
           the framer driver for the digital lines
     \param data HSS port to monitor */
/*******************************************************************/
static void *do_monitor_digital (void *data)
{

    struct port *p = data;
    unsigned int deviceID = 0;
    icp_framerdrv_event_t event;
    uint8_t state = 0;
    unsigned int lineID = 0;
    struct hss_pvt *pvt = NULL;
    struct ast_channel *chan = NULL;
    pthread_attr_t attr;
    icp_status_t status = ICP_STATUS_FAIL;

    pthread_attr_init (&attr);
    pthread_attr_setdetachstate (&attr, PTHREAD_CREATE_DETACHED);

    deviceID = p->config.portId;

    ast_log (LOG_DEBUG, "E1/T1 monitor thread started on port %d\n", deviceID);
    set_priority (MONITOR_THREAD_PRIORITY);
    
    /* Wait for events on the E1/T1 board */
    while (TRUE) {
       
        status = icp_FramerDrvRead (deviceID, &event, TRUE);

        if (status != ICP_STATUS_SUCCESS) {
            ast_log (LOG_ERROR, "Fail to read events from the framer driver\n");
        }

        ast_log (LOG_DEBUG, "I got an event on line %d timeslot %d\n", event.lineId,
                 toInt(event.ts));

        /* Compute the line ID */
        lineID = DIGITAL_LINE_MASK
                 + SLOT_MASK * deviceID
                 + SPAN_MASK * event.lineId
                 + toInt(event.ts);


        pvt = find_line (lineID);

        if (pvt == NULL) {
            ast_log (LOG_WARNING, "Got an event on an unconfigured line\n");
            continue;
        }

        switch (event.eventType) {
        case ICP_FRAMERDRV_EVENT_COSS:       
            status =
                icp_FramerDrvSignalStateGet (deviceID, event.lineId, toInt(event.ts), &state);

            if (status != ICP_STATUS_SUCCESS) {
                ast_log (LOG_ERROR,
                         "Fail to retrieve state on slot %d, span %d, timeslot %d\n",
                         pvt->slotID, pvt->spanID, pvt->timeslot);
            }

            switch (state) {
                /* ONHOOK */
            case EM_ONHOOK:
                
                ast_log (LOG_DEBUG, "Got ONHOOK on timeslot %d\n", toInt(event.ts));

                /* User hangs up while dialing (no call created) */
                if (pvt->getDigits_thread != AST_PTHREADT_NULL) {
                    /* Kill the thread reading the digits on this structure */
                    pthread_cancel (pvt->getDigits_thread);
                    pvt->getDigits_thread = AST_PTHREADT_NULL;

                    /* Bring the voice channel down */
                    BRING_CHANNEL_DOWN(pvt);

                    /* Reset extension number */
                    bzero (pvt->ext, sizeof (pvt->ext));

                }
                /* User hangs up a call */
                else if (pvt->owner[MAINCALL] != NULL) {
                    
                   chan = pvt->owner[MAINCALL];
                   
                   /* Hangup the line */
                   ast_softhangup (chan, AST_SOFTHANGUP_DEV);

                    ast_log (LOG_DEBUG, "Hanging the line up : %d\n", pvt->lineID);
                }
                break;

                /* OFFHOOK */
            case EM_OFFHOOK:

                ast_log (LOG_DEBUG, "Got OFFHOOK on timeslot %d\n", toInt(event.ts));

                /* Bring the voice channel up */
                BRING_CHANNEL_UP(pvt);

                /* User answers a call */
                if (pvt->owner[MAINCALL] != NULL) {
                    chan = pvt->owner[MAINCALL];

                    
                    if (chan->_state == AST_STATE_RINGING) {
                        
                        ast_log (LOG_DEBUG, "Answering the line\n");
                        
                        ast_setstate (chan, AST_STATE_UP);
                        ast_queue_control (chan, AST_CONTROL_ANSWER);
                        
                    } else {
                        
                        ast_log (LOG_ERROR, "Received unexpected OFFHOOK on line %d\n", pvt->lineID);                    
                    }
                    
                /*  Wait for digits. Make sure only one thread is started per timeslot */
                } else if (pvt->getDigits_thread == AST_PTHREADT_NULL){
                    
                    ast_pthread_create (&pvt->getDigits_thread, &attr, get_digits, pvt);
                }
                break;
            default:
            
                ast_log (LOG_ERROR, "Received an invalid state %d on timeslot %d\n",
                         state, toInt(event.ts));
                break;
            }

            break;

        case ICP_FRAMERDRV_EVENT_RED:
            ast_log (LOG_ERROR, "Red alarm on device %d!!\n", deviceID);
            break;

        case ICP_FRAMERDRV_EVENT_YEL:
            ast_log (LOG_ERROR, "Yellow alarm on device %d!!\n", deviceID);
            break;

        default:
            break;
        }

    }

    return NULL;
}




/***************************************************************/
/*! \brief Unload chan_hss module and release CVI components
    \return 0 on success else -1 */
/***************************************************************/
static int __unload_module (void)
{
    int i = 0;
    int voicefd = 0;            /* Descriptor to bring down the HSS ports */
    struct hss_pvt localinterface;
    struct hss_pvt *pvt;
    icp_status_t status = ICP_STATUS_FAIL;

    /* Unregister the chan technology */
    ast_channel_unregister (&hss_tech);
    
    /* Unregister the CLI command */
    for (i = 0; i < NB_HSS_PORTS; i++) {
        if ((hssPorts[i].mode == E1) || (hssPorts[i].mode == T1)){
            ast_cli_unregister (&hss_cli);
            break;
        }
    }

    if (ast_mutex_lock (&iflock)) {
        ast_log (LOG_ERROR, "Unable to lock interface list???\n");
        return -1;
    }

    /* Kill the FXO-FXS monitor thread */
    if (monitor_thread != AST_PTHREADT_NULL) {
        pthread_cancel (monitor_thread);
    }
    
    /* Kill the monitor threads on DIGITAL slots */
    for (i = 0; i < NB_HSS_PORTS; i++) {
        if (hssPorts[i].monitorThread != AST_PTHREADT_NULL) {
            pthread_cancel (hssPorts[i].monitorThread);
        }
    }
    
    /* Soft hang up all the interfaces in use */
    for (i = 0; i < MAX_CHANNELS_NUM; i++) {
        localinterface = interfaces[i];

        /* This interface has not been configured */
        if (localinterface.fd < 0) {
            continue;
        }

        if (localinterface.owner[MAINCALL] != NULL) {
            ast_softhangup (localinterface.owner[MAINCALL], AST_SOFTHANGUP_APPUNLOAD);
        }

        if (localinterface.owner[WAITINGCALL] != NULL) {
            ast_softhangup (localinterface.owner[WAITINGCALL], AST_SOFTHANGUP_APPUNLOAD);
        }

        /*if the line is up waiting for digits, kill the thread retrieving them */
        if (localinterface.getDigits_thread != AST_PTHREADT_NULL) {
            pthread_cancel (localinterface.getDigits_thread);
        }
        
        /* Destroy DSP */
        ast_dsp_free (localinterface.dsp);

#ifdef USE_OSLEC_EC
        /* Destroy echo canceller */
        if (localinterface.echocancel) {
            echo_can_free (localinterface.ec);
        }
#endif

        if (localinterface.mode == FXS) {
            /*Stop ringing and playing tones */
            if (ioctl (slicfd, ICP_ANALOGDRV_RING_OFF, localinterface.lineID)) {
                ast_log (LOG_ERROR, "Fail to stop ringing line %d\n",
                         localinterface.lineID);
            }

            if (ioctl (slicfd, ICP_ANALOGDRV_TONE_OFF, localinterface.lineID)) {
                ast_log (LOG_ERROR, "Fail to stop playing tone\n");
            }
        }

        /* Remove channel and close file descriptor */
        pvt = &localinterface;
        BRING_CHANNEL_DOWN(pvt);
        
        if (ioctl (localinterface.fd, ICP_HSSVOICEDRV_CHAN_REMOVE, 0)) {
            ast_log (LOG_ERROR, "Fail to remove the voice channel\n");
        }

        close (localinterface.fd);
    }

    ast_mutex_unlock (&iflock);

    /* If the SLIC/CODEC driver has been used, then close the file descriptor */
    if (slicfd > 0) {
        close (slicfd);
    }

    /* Open a file descriptor to bring down the HSS ports */
    voicefd = open (VOICE_DEVICE, O_RDWR);

    if (voicefd < 0) {
        ast_log (LOG_ERROR, "Can't open the voice driver to bring down the ports\n");
        return -1;
    }

    for (i = 0; i < NB_HSS_PORTS; i++) {

        if (hssPorts[i].up == TRUE) {
            if (ioctl (voicefd, ICP_HSSVOICEDRV_PORT_DOWN, i)) {
                ast_log (LOG_ERROR, "Fail to bring down HSS port %d\n", i);
            }
            hssPorts[i].up = FALSE;
        }

        if ((hssPorts[i].mode == E1) || (hssPorts[i].mode == T1)) {
            status = icp_FramerDrvUninit (i);
            

            if (status != ICP_STATUS_SUCCESS) {
                ast_log (LOG_ERROR, "Fail to uninit the Framer driver on HSS port %d\n",
                         i);
            }
        }
    }

    close (voicefd);

    return 0;
}



/***************************************************************/
/*! \brief Asterisk callback. Unload chan_hss module
    \return 0 on success else -1 */
/***************************************************************/
static int unload_module (void)
{
    return __unload_module ();
}





/***************************************************************/
/*! \brief Asterisk callback. Read configuration file and 
           initialise CVI components
    \return 0 on success else -1 */
/***************************************************************/
static int load_module (void)
{
    struct ast_config *configFile = NULL;
    struct ast_variable *variable = NULL;
    int portID = 0;
    int i = 0;
    unsigned int preconfig = 0xFF;
    static int geoconfig = US_FCC;
    port_mode mode = -1;
    port_setting setting = -1;
    
    nbInterfaces = 0;
    nbFxsFxo = 0;
    nbDigital = 0;
    geoconfig = US_FCC;

    for (portID = 0; portID < NB_HSS_PORTS; portID++) {
        hssPorts[portID].configured = FALSE;
        hssPorts[portID].up = FALSE;
        hssPorts[portID].config.portId = portID;
        hssPorts[portID].config.loopbackMode = ICP_HSSDRV_NO_LOOPBACK;
        hssPorts[portID].availableTimeslot = 0;
        hssPorts[portID].monitorThread = AST_PTHREADT_NULL;
    }


    for (i = 0; i < MAX_CHANNELS_NUM; i++) {
        /* Init file descriptor to -1, so we can detect if the interface has been configured or not */
        interfaces[i].fd = -1;
        /* init mutex */
        ast_mutex_init (&interfaces[i].lock);
    }

    /* Load the configuration file */
    configFile = ast_config_load (CONFIG_FILE);

    if (configFile == NULL) {
        ast_log (LOG_ERROR, "Fail to open config file\n");
        return -1;
    }

    if (ast_mutex_lock (&iflock)) {
        ast_log (LOG_ERROR, "Unable to lock interface list???\n");
        return -1;
    }

    /* Read the configuration section */
    variable = ast_variable_browse (configFile, "configuration");

    while (variable != NULL) {

        if (!strncasecmp (variable->name, "mode", 4)) {

            if (!strncasecmp (variable->value, "E1", 2)) {
                mode = E1;
            } else if (!strncasecmp (variable->value, "T1", 2)) {
                mode = T1;
            } else if (!strncasecmp (variable->value, "SLIC", 4)) {
                mode = SLIC;
            } else {
                ast_log (LOG_ERROR, "Unknown parameter\n");
                ast_config_destroy (configFile);
                ast_mutex_unlock (&iflock);
                unload_module ();
                return -1;
            }
        } else if (!strncasecmp (variable->name, "port", 4)) {
        
            if (*(variable->value) == '\0') {
                ast_log (LOG_ERROR, "Null value for port ID\n");
                ast_config_destroy (configFile);
                ast_mutex_unlock (&iflock);
                unload_module ();
                return -1;
            }
            
            if (configure_port (atoi (variable->value), mode, setting, preconfig, geoconfig) < 0) {
                ast_log (LOG_ERROR, "Unable to configure port\n");
                ast_config_destroy (configFile);
                ast_mutex_unlock (&iflock);
                unload_module ();
                return -1;
            }

        } else if (!strncasecmp (variable->name, "geo", 3)) {
            if (!strncasecmp (variable->value, "us", 2)) {
                geoconfig = US_FCC;
            } else if (!strncasecmp (variable->value, "eu", 2)) {
                geoconfig = EU_TBR21;
            }
            
        } else if (!strncasecmp (variable->name, "preconfig", 9)) {
            preconfig = atoi (variable->value);

        } else if (!strncasecmp (variable->name, "setting", 7)) {
        
            if (!strncasecmp (variable->value, "SINGLE", 6)) {
                setting = SINGLE;
            } else if (!strncasecmp (variable->value, "QUAD", 4)) {
                setting = QUAD;
            } else {
                ast_log (LOG_ERROR, "Unknown parameter\n");
                ast_config_destroy (configFile);
                ast_mutex_unlock (&iflock);
                unload_module ();
                return -1;
            }
        }


        variable = variable->next;
    }

    /* Read the channels section file and create each interface */
    variable = ast_variable_browse (configFile, "channels");

    while (variable != NULL) {
        if (!strncasecmp (variable->name, "line", 4)) {

            /* Build the interfaces and add them in the interfaces array */
            if (build_channels (variable->value) == -1) {
                ast_log (LOG_ERROR, "Fail to create interfaces list\n");
                ast_config_destroy (configFile);
                ast_mutex_unlock (&iflock);
                unload_module ();
                return -1;
            }

        } else if (!strncasecmp (variable->name, "callerid", 8)) {
            ast_callerid_split (variable->value, callerid_name, sizeof (callerid_name),
                                callerid_num, sizeof (callerid_num));
                                
        } else if (!strncasecmp (variable->name, "context", 7)) {
            ast_copy_string (context, variable->value, sizeof (context));
            
        } else if (!strncasecmp (variable->name, "format", 6)) {
            if (!strncasecmp (variable->value, "ulaw", 4)) {
                configFormat = AST_FORMAT_ULAW;
                
            } else if (!strncasecmp (variable->value, "alaw", 4)) {
                configFormat = AST_FORMAT_ALAW;
                
            } else if (!strncasecmp (variable->value, "linear", 6)) {
                configFormat = AST_FORMAT_SLINEAR;
            }

            /* Raise an error because the codec format is not supported */
            else {
                ast_log (LOG_ERROR, "Unknown codec format\n");
                ast_config_destroy (configFile);
                ast_mutex_unlock (&iflock);
                unload_module ();
                return -1;
            }

        }

        else if (!strncasecmp (variable->name, "silencesuppression", 18)) {
            silencesuppression = ast_true (variable->value);
            
        } else if (!strncasecmp (variable->name, "callwaiting", 11)) {
            callwaiting = ast_true (variable->value);
            
        } else if (!strncasecmp (variable->name, "echocancel", 10)) {
            echocancel = ast_true (variable->value);
        }


        variable = variable->next;
    }

    ast_config_destroy (configFile);

    ast_mutex_unlock (&iflock);



    /* Register the chan technology */
    if (ast_channel_register (&hss_tech)) {
        ast_log (LOG_ERROR, "Fail to register channel HSS\n");
        unload_module ();
        return -1;
    }
    
   
    portID = 0;

    /* If at least 1 analog board is configured then start the FXS-FXO monitor thread */
    while ((portID < NB_HSS_PORTS) && (monitor_thread == AST_PTHREADT_NULL)) {
        if ((hssPorts[portID].mode == SLIC) && (hssPorts[portID].configured == TRUE)) {
            if (ast_pthread_create_background (&monitor_thread,
                                               NULL,
                                               do_monitor,
                                               NULL) < 0) {
                ast_log (LOG_ERROR, "Fail to start monitor thread\n");
                unload_module ();
                return -1;
            }
        }

        portID++;
    }


    /* Start monitor threads on DIGITAL slots */
    for (portID = 0; portID < NB_HSS_PORTS; portID++) {
        if ((hssPorts[portID].mode == E1) || (hssPorts[portID].mode == T1)) {
            if (ast_pthread_create_background (&hssPorts[portID].monitorThread,
                                               NULL,
                                               do_monitor_digital,
                                               (struct port *) &hssPorts[portID]) < 0) {
 
                ast_log (LOG_ERROR, "Fail to start DIGITAL monitor thread on HSS port %d\n",
                         portID);
                unload_module ();
                return -1;
            }
        }
    }
    
    /* If one port is configured for framer, register CLI command */
    for (portID = 0; portID < NB_HSS_PORTS; portID++) {
        if ((hssPorts[portID].mode == E1) || (hssPorts[portID].mode == T1)) {
            if (ast_cli_register (&hss_cli)) {
                ast_log (LOG_ERROR, "Fail to register \"hss framer status\" CLI command\n");
                unload_module ();
                return -1;
            }
        break;
        }
    }

    return 0;
}


AST_MODULE_INFO_STANDARD (ASTERISK_GPL_KEY, DESCRIPTION);
